from mnspy.interpolación import Interpolacion
from mnspy.utilidades import es_notebook
from tabulate import tabulate
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

import sympy as sp

sp.init_printing(use_latex=True)


class InterpolacionNewton(Interpolacion):
    """Calcula el polinomio de interpolación de Newton usando diferencias divididas.

    Este método construye un polinomio que pasa por todos los puntos de datos.
    Es computacionalmente más eficiente que el método de Lagrange, ya que
    los coeficientes (diferencias divididas) pueden ser calculados de forma
    recursiva y almacenados en una tabla.

    Attributes
    ----------
    x : np.ndarray
        Array con los datos de la variable independiente.
    y : np.ndarray
        Array con los datos de la variable dependiente.

    Methods
    -------
    evaluar(x: float):
        Evalúa el polinomio de interpolación en un punto `x`.
    obtener_polinomio():
        Genera la expresión simbólica del polinomio de interpolación.
    graficar():
        Genera una gráfica del polinomio y los puntos de datos.

    Examples
    -------
    from mnspy import InterpolacionNewton
    import numpy as np

    x = np.array([1., 4., 6., 5.])
    y = np.log(x)
    inter = InterpolacionNewton(x, y)
    print(inter.evaluar(2))
    inter.graficar(2)
    inter.mostrar_diferencias_divididas()
    print(inter.obtener_polinomio(expandido=True))
    """
    def __init__(self, x: np.ndarray, y: np.ndarray):
        """Constructor de la clase InterpolacionNewton.

        Parameters
        ----------
        x: np.ndarray
            Array con los datos de la variable independiente
        y: np.ndarray
            Array con los datos de la variable dependiente
        """
        super().__init__(x, y)
        # Calcula la tabla de diferencias divididas
        self._b = np.zeros((self._n, self._n))
        self._b[:, 0] = np.transpose(self._y)

        for j in range(1, self._n):
            for i in range(self._n - j):
                self._b[i, j] = (self._b[i + 1, j - 1] - self._b[i, j - 1]) / (self._x[i + j] - self._x[i])

    def evaluar(self, x: float) -> float:
        """Evalúa el polinomio de Newton en un punto `x` usando la forma anidada.

        Parameters
        ----------
        x: float
            Valor de la variable independiente en el que se desea interpolar.

        Returns
        -------
        float
            El valor interpolado y(x).
        """
        x_t = 1
        y_int = self._b[0, 0]
        for j in range(self._n - 1):
            x_t *= (x - self._x[j])  # Término (x - x_0)(x - x_1)...
            y_int += self._b[0, j + 1] * x_t
        return y_int

    def obtener_polinomio(self, expandido: bool = False):
        """Genera la expresión simbólica del polinomio de interpolación.

        Parameters
        ----------
        expandido: bool
            Si es verdadero, retorna el polinomio expandido, en caso contrario muestra el polinomio en forma
            de diferencias divididas.

        Returns
        -------
        Retorna el polinomio que pasa por los puntos de los datos de la interpolación
        """
        x = sp.symbols('x')
        pol = sum([self._b[0][i] * np.prod([(x - self._x[j]) for j in range(i)]) for i in range(self._n)])
        if expandido:
            return sp.expand(pol)
        else:
            return pol

    def mostrar_diferencias_divididas(self):
        """Muestra la tabla de diferencias divididas finitas.

        Returns
        -------
        str or None
            Una tabla HTML para notebooks de Jupyter, o imprime la tabla en la consola.
        """
        tabla = {}
        if es_notebook():
            tabla['$x_{i}$'] = self._x
            tabla['$y_{i}$'] = self._y
            for i in range(self._n - 1):
                dato = list([''] * self._n)
                for j in range(self._n - 1 - i):
                    dato[j] = self._b[j, i + 1]
                tabla['$dif_{' + str(i + 1) + '}$'] = dato
            return tabulate(tabla, headers='keys', tablefmt='html')
        else:
            tabla['x_i'] = self._x
            tabla['y_i'] = self._y
            for i in range(self._n - 1):
                dato = list([''] * self._n)
                for j in range(self._n - 1 - i):
                    dato[j] = self._b[j, i + 1]
                tabla['dif_' + str(i + 1)] = dato
            print(tabulate(tabla, headers='keys', tablefmt='simple'))

    def graficar(self, x: float) -> None:
        """Genera una gráfica del polinomio de interpolación.

        Dibuja el polinomio y resalta los puntos de datos originales, así como
        el punto interpolado `(x, y(x))`.

        Parameters
        ----------
        x: float
            Valor de `x` a interpolar y resaltar en la gráfica.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        y = self.evaluar(x)
        x_min = min(self._x)
        x_max = max(self._x)
        x_list = np.linspace(x_min, x_max, 1000)
        y_list = [self.evaluar(val) for val in x_list]
        plt.scatter(x, y, c='r', lw=2, label='Interpolación Newton')
        plt.plot(x_list, y_list, linestyle='dashed', c='k', lw=1, label='Polinomio')
        plt.annotate('$(' + str(x) + r',\,' + str(y) + ')$', (x, y), c='r', alpha=0.9, textcoords="offset points",
                     xytext=(0, 10), ha='center')
        super()._graficar_datos()


def main():
    x = np.array([1., 4., 6., 5.])
    y = np.log(x)
    inter = InterpolacionNewton(x, y)
    print(inter.evaluar(2))
    inter.graficar(2)
    inter.mostrar_diferencias_divididas()


if __name__ == '__main__':
    main()



from mnspy.integrales import Integral
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)
class GaussLegendre(Integral):
    """Calcula la integral de una función usando la cuadratura de Gauss-Legendre.

    Este método aproxima la integral evaluando la función en puntos específicos
    (los nodos de Gauss) y sumando estos valores con pesos precalculados. Es
    muy preciso para el número de puntos utilizados.

    Attributes
    ----------
    f: callable
        Función a integrar.
    a: float
        Límite inferior de integración.
    b: float
        Límite superior de integración.
    integral : float
        Resultado del cálculo de la integral.

    Methods
    -------
    graficar():
        Genera una gráfica del proceso de integración.

    Examples
    -------
    from mnspy import GaussLegendre
    import numpy as np

    def g(x):
        return 0.2 + 25 * x - 200 * x ** 2 + 675 * x ** 3 - 900 * x ** 4 + 400 * x ** 5

    gl = GaussLegendre(g, 0, 0.8, n=6)
    gl.graficar()
    print(gl.integral)
    """
    def __init__(self, f: callable, a: float, b: float, n: int = 2):
        """Constructor de la clase GaussLegendre

        Parameters
        ----------
        f: callable
            Función a integrar, f(x).
        a: float
            Límite inferior de integración.
        b: float
            Límite superior de integración.
        n: int, optional
            Número de puntos a utilizar (2, 3, 4, 5 o 6). Si se ingresa
            un valor no válido, se usará n=2 por defecto.
        """
        super().__init__(f=f, a=a, b=b, n=n)
        # Coeficientes y puntos de la cuadratura de Gauss-Legendre
        if self._n == 2:
            self._c = [1, 1]
            self._x = [-0.577350269, 0.577350269]
        elif self._n == 3:
            self._c = [0.5555556, 0.8888889, 0.5555556]
            self._x = [-0.774596669, 0, 0.774596669]
        elif self._n == 4:
            self._c = [0.3478548, 0.6521452, 0.6521452, 0.3478548]
            self._x = [-0.861136312, -0.339981044, 0.339981044, 0.861136312]
        elif self._n == 5:
            self._c = [0.2369269, 0.4786287, 0.5688889, 0.4786287, 0.2369269]
            self._x = [-0.906179846, -0.538469310, 0, 0.538469310, 0.906179846]
        elif self._n == 6:
            self._c = [0.1713245, 0.3607616, 0.4679139, 0.4679139, 0.3607616, 0.1713245]
            self._x = [-0.932469514, -0.661209386, -0.238619186, 0.238619186, 0.661209386, 0.932469514]
        else:
            self._n = 2
            self._c = [1, 1]
            self._x = [-0.577350269, 0.577350269]
        self._calcular()

    def _calcular(self):
        """Calcula la integral por el método de Gauss-Legendre.

        El resultado se almacena en el atributo `self.integral`.

        Returns
        -------
        None
        """
        suma = 0
        for i in range(self._n):
            # Transformación de la variable de [-1, 1] al intervalo [a, b]
            x_transformado = 0.5 * (self._b - self._a) * self._x[i] + 0.5 * (self._b + self._a)
            suma += self._c[i] * self._f(x_transformado)

        self.integral = 0.5 * (self._b - self._a) * suma

    def graficar(self):
        """Genera una gráfica del proceso de integración.

        Muestra la función y resalta los puntos de Gauss-Legendre evaluados.
        """
        x = list((0.5 * (self._b - self._a) * x_i + 0.5 * (self._b + self._a) for x_i in self._x))
        y = list((self._f(x_i) for x_i in x))
        plt.stem(x, y, linefmt='C2--', markerfmt='C0o', basefmt='C2-', label=f'Gauss-Legendre (n={self._n})')
        plt.title(f'$\\int{{f(x)\\,dx}} \\approx {self.integral:.8f}$')
        self._graficar_datos()


def main():
    """Función principal para demostración."""
    def g(x):
        return 0.2 + 25 * x - 200 * x ** 2 + 675 * x ** 3 - 900 * x ** 4 + 400 * x ** 5

    gl = GaussLegendre(g, 0, 0.8, n=6)
    gl.graficar()
    print(gl.integral)


if __name__ == '__main__':
    main()



from mnspy.ecuaciones_diferenciales_parciales import EcuacionesDiferencialesParciales
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class Liebmann(EcuacionesDiferencialesParciales):
    """Soluciona la ecuación de Laplace en 2D usando el método de Liebmann.

    Este método es una implementación del método de Gauss-Seidel para resolver
    el sistema de ecuaciones lineales que resulta de discretizar la ecuación
    de Laplace mediante diferencias finitas. Incluye un factor de
    sobrerrelajación (lambda) para acelerar la convergencia.

    Attributes
    ----------
    converge: bool
        Indica si el método convergió dentro de la tolerancia y el número
        máximo de iteraciones.
    factor_lambda: float
        Factor de sobrerrelajación utilizado en los cálculos.
    iter: int
        Número de iteraciones realizadas.
    iter_max: int
        Máximo número de iteraciones permitidas.
    tol_porc: float
        Tolerancia de error porcentual para el criterio de parada.

    Methods
    -------
    _calcular():
        Ejecuta el algoritmo iterativo de Liebmann.
    graficar():
        Genera una gráfica de contorno de la solución.
    graficar_campos():
        Genera una gráfica de vectores para los campos de flujo.

    Examples:
    -------
    from mnspy import Liebmann

    lp = Liebmann((10, 10), {'norte': 80.0, 'sur': 20.0, 'oeste': 20.0, 'este': 0.0}, 5, tol_porc=1)
    lp.graficar()
    lp.graficar_campos()

    """
    def __init__(self, n: int | tuple[int, int], frontera: dict[str, float | str | list[float]], val_inicial: float,
                 tol_porc: float = 0.1, factor_lambda: float = 1, iter_max: int = 200, k_x: float = 1, k_y: float = 1):
        """Constructor de la clase Liebmann

        Parameters
        ----------
        n: int | tuple[int, int]
            Número de nodos en cada dirección. Si es un entero, la malla será de
            n x n. Si es una tupla (n_filas, n_columnas), la malla será de
            n_filas x n_columnas.
        frontera: dict[str, float | str | list[float]]
            Condiciones de frontera de la placa ('norte', 'sur', 'este', 'oeste').
            Los valores pueden ser de tipo ``float`` (Dirichlet), ``list[float]`` (Dirichlet)
            o ``'aislado'`` (Neumann).
        val_inicial: float
            Valor inicial para todos los nodos interiores de la malla.
        tol_porc: float, optional
            Tolerancia porcentual de error para el criterio de parada. Por defecto es 0.1.
        factor_lambda: float, optional
            Factor de sobrerrelajación (1 <= lambda < 2). Por defecto es 1 (Gauss-Seidel).
        iter_max: int, optional
            Número máximo de iteraciones. Por defecto es 200.
        k_x: float, optional
            Coeficiente de conductividad térmica en la dirección x. Por defecto es 1.
        k_y: float, optional
            Coeficiente de conductividad térmica en la dirección y. Por defecto es 1.
        """
        super().__init__(n, frontera, val_inicial, k_x, k_y)
        self.iter_max = iter_max
        self.factor_lambda = factor_lambda
        self.tol_porc = tol_porc
        self.converge = True
        self.iter = 0
        self._calcular()

    def _calcular(self):
        """Ejecuta los cálculos iterativos del método de Liebmann.
        """
        for k in range(self.iter_max):
            self.iter = k + 1
            self.converge = True
            # Manejo de fronteras de Neumann (aisladas)
            if self._frontera['norte'] == 'aislado':
                self.U[self._n - 1, 1:self._m - 1] = (self.U[self._n - 2, 1:self._m - 1] / 2 +
                                                       self.U[self._n - 1, :self._m - 2] / 4 + self.U[self._n - 1,
                                                                                          2:self._m] / 4)
                self.q_x[self._n - 1, 1:self._m - 1] = -(self.U[self._n - 1, 2:self._m] - self.U[self._n - 1,
                                                                                      :self._m - 2]) / 2
            if self._frontera['sur'] == 'aislado':
                self.U[0, 1:self._m - 1] = (self.U[1, 1:self._m - 1] / 2 + self.U[0, :self._m - 2] / 4 +
                                             self.U[0, 2:self._m] / 4)
                self.q_x[0, 1:self._m - 1] = -(self.U[0, 2:self._m] - self.U[0, :self._m - 2]) / 2

            if self._frontera['oeste'] == 'aislado':
                self.U[1:self._n - 1, 0] = (
                        self.U[1:self._n - 1, 1] / 2 + self.U[2:self._n, 0] / 4 + self.U[:self._n - 2, 0] / 4)
                self.q_y[1:self._n - 1, 0] = -(self.U[2:self._n, 0] - self.U[:self._n - 2, 0]) / 2

            if self._frontera['este'] == 'aislado':
                self.U[1:self._n - 1, self._m - 1] = (
                        self.U[1:self._n - 1, self._m - 2] / 2 + self.U[2:self._n, self._m - 1] / 4 + self.U[
                                                                                                  :self._n - 2,
                                                                                                        self._m - 1] / 4)
                self.q_y[1:self._n - 1, self._m - 1] = -(self.U[2:self._n, self._m - 1] - self.U[:self._n - 2,
                                                                                            self._m - 1]) / 2
            # Itera sobre los nodos interiores
            for i in range(1, self._n - 1):
                for j in range(1, self._m - 1):
                    last_val = self.U[i, j]
                    # Fórmula de Liebmann (Gauss-Seidel con sobrerrelajación)
                    self.U[i, j] = self.factor_lambda * (
                            self.U[i + 1, j] + self.U[i - 1, j] + self.U[i, j + 1] + self.U[i, j - 1]) / 4 + (
                                           1 - self.factor_lambda) * last_val
                    # Verifica la convergencia
                    if self.U[i, j] != 0.0:
                        cumple = abs((self.U[i, j] - last_val) / self.U[i, j]) < self.tol_porc / 100.0
                    else:
                        cumple = False
                    self.converge = self.converge and cumple
            if self.converge:
                break
        self._calcular_campos()

    def graficar(self):
        """Genera una gráfica de contorno de la solución.

        Returns
        -------
        Gráfica de los resultados interpolados usando el pquete matplotlib
        """
        plt.axes().set_aspect('equal')
        plt.suptitle('Método de Liebmann')
        if self.converge:
            plt.title(f'Tolerancia= {self.tol_porc} %, N iteraciones= {self.iter}')
        else:
            plt.title(f'No converge, N iteraciones= {self.iter}')
        super()._graficar_datos()

    def graficar_campos(self):
        """Genera una gráfica de vectores para los campos de flujo.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib (quiver plot).
        """
        plt.axes().set_aspect('equal')
        plt.suptitle('Método de Liebmann')
        super()._graficar_campos()


def main():
    """Función principal para demostración."""
    lp = Liebmann((10, 10), {'norte': 80.0, 'sur': 20.0, 'oeste': 20.0, 'este': 0.0}, 5, tol_porc=1)
    lp.graficar()
    lp.graficar_campos()


if __name__ == '__main__':
    main()



from mnspy.utilidades import es_notebook, _generar_matrix, _formato_float_latex
import numpy as np
from IPython.display import display, Math
import sympy as sp
from enum import Enum
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from matplotlib import tri

TOL_CERO = 1E-10
FORMATO_NUM = '{:.10g}'
sp.init_printing(use_latex=True)


class Metodo(Enum):
    """Define los esquemas de discretización para los términos de convección.

    - **CDS (Central Differencing Scheme)**: Esquema centrado, de segundo orden
      pero puede generar oscilaciones no físicas con altos números de Péclet.
    - **UDS (Upwind Differencing Scheme)**: Esquema de diferencias contra el
      viento, de primer orden, muy estable pero introduce difusión numérica.
    - **HDS (Hybrid Differencing Scheme)**: Esquema híbrido que combina CDS y
      UDS, utilizando UDS cuando la convección es dominante (Pe > 2).
    """
    CDS = 1  # Central Differencing Scheme
    UDS = 2  # Upwind Differencing Scheme
    HDS = 3  # Hybrid Differencing Scheme


#
# class Escalar:
#     def __init__(self, variable: str, valor: float = 0.0):
#         self.variable = variable
#         self.valor = valor


class Vertice:
    """Representa un vértice (o nodo) en una malla de volúmenes finitos.

    Un vértice tiene coordenadas espaciales (x, y, z) y puede almacenar
    propiedades escalares (ej. temperatura) y vectoriales (ej. velocidad).
    """
    def __init__(self, nombre: str, x: float = 0.0, y: float = 0.0, z: float = 0.0):
        """Constructor de la clase Vertice.

        Parameters
        ----------
        nombre : str
            Identificador único para el vértice.
        x : float, optional
            Coordenada x del vértice. Por defecto es 0.0.
        y : float, optional
            Coordenada y del vértice. Por defecto es 0.0.
        z : float, optional
            Coordenada z del vértice. Por defecto es 0.0.
        """
        self.nombre = nombre
        self.x = x
        self.y = y
        self.z = z
        self.campos = None
        self.escalares = None
        self.lista_superficies = list()

    def get_punto(self) -> tuple[float, float, float]:
        """Retorna las coordenadas del vértice como una tupla."""
        return self.x, self.y, self.z

    def set_punto(self, punto: tuple[float, float, float]):
        """Establece las coordenadas del vértice."""
        self.x, self.y, self.z = punto

    def get_distancia(self, p: tuple[float, float, float]) -> float:
        """Calcula la distancia euclidiana a otro punto."""
        return float(np.sqrt((self.x - p[0]) ** 2 + (self.y - p[1]) ** 2 + (self.z - p[2]) ** 2))

    def set_escalar(self, var: dict):
        """Agrega o actualiza un valor escalar asociado al vértice.

        Parameters
        ----------
        var : dict
            Diccionario con el nombre de la variable y su valor.
            Ejemplo: `{'T': 300.0}`.
        """
        if self.escalares is None:
            self.escalares = var
        else:
            self.escalares.update(var)

    def get_escalar(self) -> dict['str': float | list | np.ndarray]:
        """Retorna el diccionario de escalares del vértice."""
        return self.escalares

    def set_campo(self, var: dict[str, list | np.ndarray]):
        """Agrega o actualiza un campo vectorial asociado al vértice.

        Parameters
        ----------
        var : dict[str, list | np.ndarray]
            Diccionario con el nombre del campo y su valor vectorial.
            Ejemplo: `{'velocidad': [1.0, 0.5, 0.0]}`.
        """
        if self.campos is None:
            self.campos = var
        else:
            self.campos.update(var)

    def get_campo(self) -> dict[str, list | np.ndarray]:
        """Retorna el diccionario de campos vectoriales del vértice."""
        return self.campos


class Superficie:
    """Representa una superficie (o cara) que delimita un volumen de control (Celda).

    Una superficie está definida por una lista de vértices y tiene propiedades
    geométricas como área y vector normal. Actúa como interfaz entre dos
    celdas vecinas (o una celda y una frontera del dominio).

    Attributes
    ----------
    nombre : str
        Identificador único para la superficie.
    lista_vertices : list[int]
        Lista de índices de los vértices que definen la superficie.
    t : float
        Espesor (para superficies 2D representadas como líneas).
    u : tuple[float, float, float]
        Vector de velocidad en la superficie, utilizado en problemas de convección.
    A : float
        Área de la superficie.
    centroide : Vertice
        Objeto `Vertice` que representa el centroide de la superficie.
    normal : np.ndarray
        Vector normal unitario a la superficie.
    """
    def __init__(self, nombre: str, vertices: list[int], t: float = 1.0,
                 u: tuple[float, float, float] = (0.0, 0.0, 0.0)):
        """Constructor de la clase Superficie.

        Parameters
        ----------
        nombre : str
            Identificador único para la superficie.
        vertices : list[int]
            Lista de índices de los vértices que la componen.
        t : float, optional
            Espesor, para cálculos de área en 2D. Por defecto es 1.0.
        u : tuple[float, float, float], optional
            Vector de velocidad en la cara. Por defecto es (0,0,0).
        """
        self.nombre = nombre
        self.lista_vertices = vertices
        self.vertices = None
        self.t = t
        self.A = None
        self.u = u
        self._vecinos = [None, None]
        self.centroide = Vertice(nombre)
        self.normal = None

    def set_vertices(self, vertices: list[Vertice]):
        """Asigna los objetos `Vertice` a la superficie."""
        self.vertices = vertices

    def set_centroide(self, centroide: tuple[float, float, float]):
        """Establece las coordenadas del centroide de la superficie."""
        self.centroide.set_punto(centroide)

    def set_normal(self, normal: np.ndarray):
        """Establece el vector normal de la superficie."""
        self.normal = normal

    def get_vecinos(self) -> list:
        """Retorna la lista de celdas vecinas a esta superficie."""
        return self._vecinos

    def get_vecino(self, vecino: any):
        """Obtiene la celda vecina opuesta a la celda dada.

        Parameters
        ----------
        vecino : Celda
            La celda de referencia.

        Returns
        -------
        Celda or None
            La celda vecina, o `None` si es una frontera.
        """
        if vecino in self.get_vecinos():
            for c in self.get_vecinos():
                if c != vecino:
                    return c
        else:
            return None

    def set_vecino(self, vecino: any):
        """Asigna una celda como vecina a esta superficie.

        Una superficie puede tener como máximo dos celdas vecinas.
        """
        if self._vecinos[0] is None:
            self._vecinos[0] = vecino
        elif self._vecinos[1] is None:
            self._vecinos[1] = vecino
        else:
            raise IndexError("Una superficie no puede tener más de dos celdas vecinas.")


class SuperficieDirichlet(Superficie):
    """Representa una condición de frontera de tipo Dirichlet (valor fijo).

    Hereda de `Superficie` y añade un atributo para almacenar el valor
    fijo de la variable en la frontera (ej. temperatura fija).
    """
    def __init__(self, nombre: str, vertices: list[int], var: dict[str:float] = None, t: float = 1.0,
                 u: tuple[float, float, float] = (0.0, 0.0, 0.0)):
        """Constructor para una superficie de frontera Dirichlet.

        Parameters
        ----------
        var : dict[str:float], optional
            Diccionario que contiene el nombre de la variable y su valor fijo.
            Ejemplo: `{'T': 300.0}`. Por defecto es `{'T': 0.0}`.
        """
        super().__init__(nombre, vertices, t, u)
        if var is None:
            self.var = {'T': 0.0}
        else:
            self.var = var


class SuperficieNeumann(Superficie):
    """Representa una condición de frontera de tipo Neumann (flujo fijo).

    Hereda de `Superficie` y añade un atributo para especificar el flujo
    (gradiente de la variable) a través de la frontera. Un flujo de cero
    representa una frontera aislada o de simetría.
    """
    def __init__(self, nombre: str, vertices: list[int], flujo: float = 0.0, t: float = 1.0,
                 u: tuple[float, float, float] = (0.0, 0.0, 0.0)):
        """Constructor para una superficie de frontera Neumann.

        Parameters
        ----------
        flujo : float, optional
            Valor del flujo a través de la superficie. Por defecto es 0.0 (aislada).
        """
        super().__init__(nombre, vertices, t, u)
        self.flujo = flujo


class SuperficieRobin(Superficie):
    """Representa una condición de frontera de tipo Robin (mixta).

    Hereda de `Superficie` y modela una condición que depende tanto del valor
    de la variable en la frontera como de su flujo, comúnmente usada para
    representar transferencia de calor por convección.
    """
    def __init__(self, nombre: str, vertices: list[int], h: float = 0.0, T_inf: float = 0.0, t: float = 1.0,
                 u: tuple[float, float, float] = (0.0, 0.0, 0.0)):
        """Constructor para una superficie de frontera Robin.

        Parameters
        ----------
        h : float, optional
            Coeficiente de transferencia (ej. coeficiente de convección).
        T_inf : float, optional
            Valor de referencia del ambiente (ej. temperatura del fluido circundante).
        """
        super().__init__(nombre, vertices, t, u)
        self.h = h
        self.T_inf = T_inf


def es_superficie_dirichlet(sup: Superficie) -> bool:
    """Método estático para identificar si es una superficie Dirichlet

    Parameters
    ----------
    sup: Superficie
        Objeto de tipo Superficie

    Returns
    -------
    True: Si la superficie es de tipo Dirichlet
    False: Si la superficie no es de tipo Dirichlet
    """
    return type(sup).__name__ == 'SuperficieDirichlet'


def es_superficie_neumann(sup: Superficie) -> bool:
    """Método estático para identificar si es una superficie Neumann

    Parameters
    ----------
    sup: Superficie
        Objeto de tipo Superficie

    Returns
    -------
    True: Si la superficie es de tipo Neumann
    False: Si la superficie no es de tipo Neumann
    """
    return type(sup).__name__ == 'SuperficieNeumann'


def es_superficie_robin(sup: Superficie) -> bool:
    """Método estático para identificar si es una superficie Robin

    Parameters
    ----------
    sup: Superficie
        Objeto de tipo Superficie

    Returns
    -------
    True: Si la superficie es de tipo Robin
    False: Si la superficie no es de tipo Robin
    """
    return type(sup).__name__ == 'SuperficieRobin'


class Celda:
    """Representa un volumen de control o celda, la unidad fundamental en el MVF.

    La ecuación de conservación se aplica a esta celda, resultando en una
    ecuación algebraica que relaciona el valor de la variable en el centroide
    de la celda con los valores en las celdas vecinas.

    Attributes
    ----------
    nombre : str
        Identificador único para la celda.
    gamma : float
        Coeficiente de difusión (ej. conductividad térmica).
    q : float
        Término fuente volumétrico.
    phi_p : float
        Coeficiente central (a_P) en la ecuación discretizada.
    s_u : float
        Componente constante del término fuente linealizado (S_u).

    """
    def __init__(self, nombre: str, superficies: list[int], gamma: float = 1.0, q: float = 0.0, h: float = 0.0,
                 p: float = 1.0, T_inf: float = 0.0, densidad: float = 1.0):
        """Constructor de la clase Celda.

        Parameters
        ----------
        nombre : str
            Identificador único de la celda.
        superficies : list[int]
            Lista de índices de las superficies que componen la celda.
        gamma : float, optional
            Coeficiente de difusión. Por defecto es 1.0.
        q : float, optional
            Término fuente volumétrico. Por defecto es 0.0.
        h : float, optional
            Coeficiente de transferencia (para fuentes lineales). Por defecto es 0.0.
        p : float, optional
            Perímetro (para fuentes lineales). Por defecto es 1.0.
        T_inf : float, optional
            Temperatura de referencia (para fuentes lineales). Por defecto es 0.0.
        densidad : float, optional
            Densidad del fluido (para problemas de convección). Por defecto es 1.0.
        """
        self.nombre = nombre
        self.lista_superficies = superficies
        self.superficies = None
        self.gamma = gamma
        self.q = q
        self.h = h
        self.p = p
        self.T_inf = T_inf
        self.densidad = densidad
        self.centroide = Vertice(nombre)
        self.coeficientes = dict()
        self.flujos = dict()
        self.signo_normal = None
        self.s_u = 0.0
        self.phi_p = 0.0

    def set_superficies(self, superficies: list[Superficie]):
        """Asigna los objetos `Superficie` a la celda."""
        self.superficies = superficies

    def set_centroide(self, centroide: tuple[float, float, float]):
        """Establece las coordenadas del centroide de la celda."""
        self.centroide.set_punto(centroide)

    def set_signo_normal(self, normal: list[float]):
        """Asigna los signos de las normales de las superficies respecto a la celda."""
        self.signo_normal = normal

    def calcular_ecuaciones(self, metodo: Metodo = Metodo.CDS):
        """Construye la ecuación algebraica discretizada para la celda.

        Calcula los coeficientes de la ecuación `a_P * phi_P = sum(a_nb * phi_nb) + S_u`,
        donde `phi_P` es el valor en el centroide de la celda actual, y `phi_nb`
        son los valores en las celdas vecinas. El término fuente se linealiza
        como `S = S_u + S_p * phi_P`.

        Parameters
        ----------
        metodo : Metodo, optional
            El esquema de discretización a utilizar (CDS, UDS, o HDS).
            Por defecto es `Metodo.CDS`.
        """
        if self.superficies is None:
            print('Celda no está calculada')
            return
        s_p = 0.0
        self.s_u = 0.0
        if metodo == Metodo.CDS:
            for i, s in enumerate(self.superficies):
                # Términos fuente lineales (ej. por convección perimetral)
                self.s_u += self.h * self.p * self.centroide.get_distancia(s.centroide.get_punto()) * self.T_inf
                s_p += -self.h * self.p * self.centroide.get_distancia(s.centroide.get_punto())
                # Término fuente volumétrico
                self.s_u += self.q * s.A * self.centroide.get_distancia(s.centroide.get_punto())

                # --- Contribución de cada superficie ---
                if es_superficie_dirichlet(s):
                    # Frontera de valor fijo (Dirichlet)
                    # Término difusivo
                    s_p += -self.gamma * s.A / self.centroide.get_distancia(s.centroide.get_punto())
                    self.s_u += self.gamma * s.A / self.centroide.get_distancia(s.centroide.get_punto()) * s.var['T']
                    # Convectivo
                    s_p += self.densidad * np.dot(s.u, s.normal) * self.signo_normal[i]
                    self.s_u += -self.densidad * np.dot(s.u, s.normal) * self.signo_normal[i] * s.var['T']

                elif es_superficie_neumann(s):
                    # Frontera de flujo fijo (Neumann)
                    self.s_u += s.flujo * s.A
                elif es_superficie_robin(s):
                    # Frontera mixta (Robin)
                    R_eq = s.A / (self.centroide.get_distancia(s.centroide.get_punto()) / self.gamma + 1.0 / s.h)
                    s_p += -R_eq
                    self.s_u += R_eq * s.T_inf
                else:
                    # Superficie interna (conecta con otra celda)
                    vecino = s.get_vecino(self)
                    if vecino is not None:
                        # Término difusivo
                        # El coeficiente `a_nb` se añade al diccionario de coeficientes.
                        if self.gamma == vecino.gamma:
                            # Material homogéneo
                            self.coeficientes[vecino.nombre] = self.gamma * s.A / self.centroide.get_distancia(
                                vecino.centroide.get_punto())
                        else:
                            # Material no homogéneo, se usa media armónica para la conductividad.
                            g = self.centroide.get_distancia(s.centroide.get_punto()) / self.centroide.get_distancia(
                                vecino.centroide.get_punto())
                            k = self.gamma * vecino.gamma / ((1 - g) * self.gamma + g * vecino.gamma)
                            self.coeficientes[vecino.nombre] = k * s.A / self.centroide.get_distancia(
                                vecino.centroide.get_punto())
                        # Término convectivo (CDS)
                        self.coeficientes[vecino.nombre] += -self.densidad * np.dot(s.u, s.normal) * self.signo_normal[
                            i] / 2.0
        elif metodo == Metodo.UDS:
            for i, s in enumerate(self.superficies):
                self.s_u += self.h * self.p * self.centroide.get_distancia(s.centroide.get_punto()) * self.T_inf
                s_p += -self.h * self.p * self.centroide.get_distancia(s.centroide.get_punto())
                self.s_u += self.q * s.A * self.centroide.get_distancia(s.centroide.get_punto())
                if es_superficie_dirichlet(s):
                    # Difusivo
                    s_p += -self.gamma * s.A / self.centroide.get_distancia(s.centroide.get_punto())
                    self.s_u += self.gamma * s.A / self.centroide.get_distancia(s.centroide.get_punto()) * s.var['T']
                    # Convectivo
                    s_p += -max(0.0, -self.densidad * np.dot(s.u, s.normal) * self.signo_normal[i])
                    self.s_u += max(0.0, -self.densidad * np.dot(s.u, s.normal) * self.signo_normal[i]) * s.var['T']
                elif es_superficie_neumann(s):
                    # Convectivo
                    self.s_u += s.flujo * s.A
                elif es_superficie_robin(s):
                    # Difusivo
                    R_eq = s.A / (self.centroide.get_distancia(s.centroide.get_punto()) / self.gamma + 1.0 / s.h)
                    s_p += -R_eq
                    self.s_u += R_eq * s.T_inf
                else:
                    vecino = s.get_vecino(self)
                    if vecino is not None:
                        # Difusivo
                        if self.gamma == vecino.gamma:
                            self.coeficientes[vecino.nombre] = self.gamma * s.A / self.centroide.get_distancia(
                                vecino.centroide.get_punto())
                        else:
                            g = self.centroide.get_distancia(s.centroide.get_punto()) / self.centroide.get_distancia(
                                vecino.centroide.get_punto())
                            k = self.gamma * vecino.gamma / ((1 - g) * self.gamma + g * vecino.gamma)
                            self.coeficientes[vecino.nombre] = k * s.A / self.centroide.get_distancia(
                                vecino.centroide.get_punto())
                        # Convectivo
                        self.coeficientes[vecino.nombre] += max(0.0, -self.densidad * np.dot(s.u, s.normal) *
                                                                self.signo_normal[i])
        elif metodo == Metodo.HDS:
            for i, s in enumerate(self.superficies):
                self.s_u += self.h * self.p * self.centroide.get_distancia(s.centroide.get_punto()) * self.T_inf
                s_p += -self.h * self.p * self.centroide.get_distancia(s.centroide.get_punto())
                self.s_u += self.q * s.A * self.centroide.get_distancia(s.centroide.get_punto())
                if es_superficie_dirichlet(s):
                    # Difusivo
                    s_p += -self.gamma * s.A / self.centroide.get_distancia(s.centroide.get_punto())
                    self.s_u += self.gamma * s.A / self.centroide.get_distancia(s.centroide.get_punto()) * s.var['T']
                    # Convectivo
                    s_p += -max(0.0, -self.densidad * np.dot(s.u, s.normal) * self.signo_normal[i])
                    self.s_u += max(0.0, -self.densidad * np.dot(s.u, s.normal) * self.signo_normal[i]) * s.var['T']
                elif es_superficie_neumann(s):
                    # Convectivo
                    self.s_u += s.flujo * s.A
                elif es_superficie_robin(s):
                    # Difusivo
                    R_eq = s.A / (self.centroide.get_distancia(s.centroide.get_punto()) / self.gamma + 1.0 / s.h)
                    s_p += -R_eq
                    self.s_u += R_eq * s.T_inf
                else:
                    vecino = s.get_vecino(self)
                    if vecino is not None:
                        if self.gamma == vecino.gamma:
                            self.coeficientes[vecino.nombre] = max(0.0, -self.densidad * np.dot(s.u, s.normal) * \
                                                                   self.signo_normal[i],
                                                                   self.gamma * s.A / self.centroide.get_distancia(
                                                                       vecino.centroide.get_punto()) - self.densidad * np.dot( \
                                                                       s.u, s.normal) * \
                                                                   self.signo_normal[i] / 2.0)
                        else:
                            g = self.centroide.get_distancia(s.centroide.get_punto()) / self.centroide.get_distancia( \
                                vecino.centroide.get_punto())
                            k = self.gamma * vecino.gamma / ((1 - g) * self.gamma + g * vecino.gamma)
                            self.coeficientes[vecino.nombre] = max(0.0, -self.densidad * np.dot(s.u, s.normal) *
                                                                   self.signo_normal[i],
                                                                   k * s.A / self.centroide.get_distancia(
                                                                       vecino.centroide.get_punto()) - self.densidad * np.dot(
                                                                       s.u, s.normal) *
                                                                   self.signo_normal[i] / 2.0)
        # El coeficiente central a_P se calcula como la suma de los coeficientes de los vecinos
        # menos la parte del término fuente que depende de phi_P.
        self.phi_p = sum(self.coeficientes.values()) - s_p

    def get_ecuacion(self, escalar: str = r'\phi'):
        """Retorna la ecuación de la celda en formato simbólico (SymPy)."""
        if len(self.coeficientes) == 0:
            return sp.Symbol('')
        else:
            return sp.Eq(self.phi_p * sp.Symbol(escalar + '_{' + self.nombre + '}') + sum(
                [-self.coeficientes[c] * sp.Symbol(escalar + '_{' + c + '}') for c in self.coeficientes.keys()]),
                         self.s_u)

    def diagrama_balance(self, mostrar_etiquetas: bool = True):
        """Dibuja un diagrama de la celda, mostrando los flujos a través de sus superficies.

        Útil para verificar visualmente el balance de flujos en un volumen de control.
        """
        fig, ax = plt.subplots(figsize=(15, 5))
        lista_x = []
        lista_y = []
        if len(self.lista_superficies) == 2:
            x_e = self.superficies[0].vertices[0].x
            y_e = self.superficies[0].vertices[0].y
            x_w = self.superficies[1].vertices[0].x
            y_w = self.superficies[1].vertices[0].y
            ancho = abs(x_e - x_w) / 50
            lista_x.append(x_e)
            lista_y.append(y_e)
            lista_x.append(x_w)
            lista_y.append(y_w)
            lista_p = np.array([[x_e, y_e + ancho], [x_w, y_w + ancho], [x_w, y_w - ancho], [x_e, y_e - ancho]])
            ax.add_patch(
                Polygon(lista_p, closed=True, edgecolor='lightsteelblue', facecolor='lightsteelblue', linewidth=0.5))
            if es_superficie_dirichlet(self.superficies[0]) or es_superficie_neumann(
                    self.superficies[0]) or es_superficie_robin(self.superficies[0]):
                plt.plot([x_e, x_e], [y_e - ancho, y_e + ancho], ls='--', lw=0.5, color='royalblue')
            else:
                plt.plot([x_e, x_e], [y_e - ancho, y_e + ancho], lw=0.5, color='royalblue')
            if es_superficie_dirichlet(self.superficies[1]) or es_superficie_neumann(
                    self.superficies[1]) or es_superficie_robin(self.superficies[1]):
                plt.plot([x_w, x_w], [y_e - ancho, y_e + ancho], ls='--', lw=0.5, color='royalblue')
            else:
                plt.plot([x_w, x_w], [y_e - ancho, y_e + ancho], lw=0.5, color='royalblue')

            if self.h != 0:
                q = self.h * self.p * abs(x_e - x_w) * (self.centroide.get_escalar()['T'] - self.T_inf)
                x_h = np.linspace(min(x_e, x_w), max(x_e, x_w), 50)
                if q > 0.0:
                    sentido = '<-'
                else:
                    sentido = '->'
                for x in x_h:
                    ax.annotate('', xy=(x, ancho), xycoords='data',
                                xytext=(0, 50), textcoords='offset points', va='center', ha='center', size=8,
                                arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3", color='blue', ls='--',
                                                lw=0.5))
                    ax.annotate('', xy=(x, -ancho), xycoords='data',
                                xytext=(0, -50), textcoords='offset points', va='center', ha='center', size=8,
                                arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3", color='blue', ls='--',
                                                lw=0.5))
                ax.annotate('$' + 'q' + '=' + '{:G}$'.format(q),
                            xy=(0.5 * (x_e + x_w), 0.5 * (y_e + y_w) + ancho), xycoords='data',
                            xytext=(0, 50), textcoords='offset points', va='center', ha='center', size=8)

            if self.q != 0:
                q = self.q * 0.5 * (self.superficies[0].A + self.superficies[1].A) * abs(x_e - x_w)
                ax.annotate('$' + 'q' + '=' + '{:G}$'.format(q),
                            xy=(self.centroide.x, self.centroide.y), xycoords='data',
                            xytext=(0, 0), textcoords='offset points', va='center', ha='right', size=8)
        else:
            for s in self.superficies:
                lista_x.append(s.centroide.get_punto()[0])
                lista_y.append(s.centroide.get_punto()[1])
                lista_vx = []
                lista_vy = []
                lista_p = [self.centroide.get_punto()[:2]]
                for v in s.vertices:
                    lista_vx.append(v.get_punto()[0])
                    lista_vy.append(v.get_punto()[1])
                    lista_p.append(v.get_punto()[:2])
                ax.add_patch(
                    Polygon(np.array(lista_p), closed=True, edgecolor='lightsteelblue', facecolor='lightsteelblue',
                            linewidth=0.5))
                if es_superficie_dirichlet(s) or es_superficie_neumann(s) or es_superficie_robin(s):
                    plt.plot(lista_vx, lista_vy, ls='--', lw=0.5, color='royalblue')
                else:
                    plt.plot(lista_vx, lista_vy, lw=0.5, color='royalblue')
        plt.scatter(lista_x, lista_y, color='gray', marker='.')
        lista_x = []
        lista_y = []
        lista_x.append(self.centroide.get_punto()[0])
        lista_y.append(self.centroide.get_punto()[1])
        plt.scatter(lista_x, lista_y, color='r', marker='.')
        if mostrar_etiquetas:
            ax.text(self.centroide.get_punto()[0], self.centroide.get_punto()[1], self.nombre, ha="left", va="top",
                    size=6,
                    alpha=0.8,
                    bbox=dict(boxstyle="circle,pad=0.1", fc="lightgreen", ec="darkslategrey", lw=0.2, alpha=0.7))
            for s in self.superficies:
                ax.text(s.centroide.get_punto()[0], s.centroide.get_punto()[1], s.nombre, ha="left", va="top", size=6,
                        alpha=0.8,
                        bbox=dict(boxstyle="round,pad=0.1", fc="lightblue", ec="darkslategrey", lw=0.2, alpha=0.7))
                for v in s.vertices:
                    ax.text(v.get_punto()[0], v.get_punto()[1], v.nombre, ha="left", va="bottom", size=6, alpha=0.8)
        for i, s in enumerate(self.superficies):
            pos = self.signo_normal[i] * s.normal * 50
            if self.flujos[s.nombre] > 0.0:
                sentido = '<-'
            else:
                sentido = '->'
            ax.annotate('$' + 'q' + '=' + '{:G}$'.format(abs(self.flujos[s.nombre])),
                        xy=(s.centroide.get_punto()[0:2]), xycoords='data',
                        xytext=(pos[0], pos[1]), textcoords='offset points', va='center', ha='center', size=8,
                        arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3", color='blue', lw=0.5))

        ax.axis('equal')
        ax.set_axis_off()
        # plt.ylabel('$y$')
        # plt.grid()
        # plt.title('Balance')
        # plt.xlabel('$x$')
        ax.set_xmargin(0.55)
        ax.set_ymargin(0.55)
        plt.show()



from mnspy.ecuaciones_diferenciales_parciales import EcuacionesDiferencialesParciales
from mnspy.ecuaciones_algebraicas_lineales import Gauss
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class DiferenciasFinitas(EcuacionesDiferencialesParciales):
    """Soluciona la ecuación de Laplace en 2D usando el método de Diferencias Finitas.

    Este método discretiza la ecuación de Laplace en una malla rectangular,
    generando un sistema de ecuaciones algebraicas lineales que se resuelve
    directamente mediante la eliminación de Gauss.

    Attributes
    ----------
    sel: Gauss
        Objeto `Gauss` utilizado para resolver el sistema de ecuaciones lineales.

    Methods
    -------
    _calcular():
        Construye y resuelve el sistema de ecuaciones lineales.
    graficar():
        Genera una gráfica de contorno de la solución.
    graficar_campos():
        Genera una gráfica de vectores para los campos de flujo.

    Examples:
    -------
    from mnspy import DiferenciasFinitas

    df = DiferenciasFinitas((5, 5), {'norte': 100.0, 'sur': 0.0, 'oeste': 75.0, 'este': 50.0}, k_x=2.0)
    df.graficar()
    df.graficar_valores()
    df.graficar_coordenadas()
    """
    def __init__(self, n: int | tuple[int, int], frontera: dict[str, float | str | list[float]], k_x: float = 1.0,
                 k_y: float = 1.0):
        """Constructor de la clase DiferenciasFinitas.

        Parameters
        ----------
        n: int | tuple[int, int]
            Número de divisiones que tendrá la placa, si se ingresa un entero n la placa estará dividida en n x n.
            En caso de que se ingrese una tupla (n, m), la placa estara dividida en n x m
        frontera: dict[str, float | str | list[float]]
            Condiciones de frontera ('norte', 'sur', 'este', 'oeste').
            Los valores pueden ser de tipo ``float`` (Dirichlet), ``list[float]`` (Dirichlet)
            o ``'aislado'`` (Neumann).
        k_x: float, optional
            Coeficiente de conductividad térmica en la dirección x. Por defecto es 1.
        k_y: float, optional
            Coeficiente de conductividad térmica en la dirección y. Por defecto es 1.
        """
        super().__init__(n, frontera, 0.0, k_x, k_y)
        self.sel = None
        self._calcular()

    def _calcular(self):
        """Construye y resuelve el sistema de ecuaciones [A]{x} = {b}.
        """
        # Determina el tamaño del sistema basado en los nodos interiores
        n = self._n - 2
        m = self._m - 2
        if self._frontera['sur'] == 'aislado':
            n += 1
        if self._frontera['norte'] == 'aislado':
            n += 1
        if self._frontera['oeste'] == 'aislado':
            m += 1
        if self._frontera['este'] == 'aislado':
            m += 1
        A = np.zeros((n * m, n * m))

        # Llena la matriz A con el stencil de diferencias finitas para Laplace (5 puntos)
        for i in range(n * m):
            for j in range(n * m):
                if i == j:
                    A[i, j] = 4
                elif j == i + m:
                    A[i, j] = -1
                elif i == j + m:
                    A[i, j] = -1
                elif j == i + 1 and j % m:
                    A[i, j] = -1
                elif j == i - 1 and i % m:
                    A[i, j] = -1

        # Modifica la matriz A y el vector b para las condiciones de frontera
        b = np.zeros(n * m)
        ini_i = ini_j = 1
        if self._frontera['sur'] == 'aislado':
            ini_j = 0
            for i in range(m):
                A[i, m + i] -= 1
        else:
            b[:m] += self._frontera['sur']
        if self._frontera['norte'] == 'aislado':
            for i in range(m):
                A[m * (n - 1) + i, m * (n - 2) + i] -= 1
        else:
            b[m * (n - 1):] += self._frontera['norte']
        if self._frontera['oeste'] == 'aislado':
            ini_i = 0
            for i in range(n):
                A[m * i, m * i + 1] -= 1
        else:
            b[::m] += self._frontera['oeste']
        if self._frontera['este'] == 'aislado':
            for i in range(n):
                A[m * (i + 1) - 1, m * (i + 1) - 2] -= 1
        else:
            b[m - 1::m] += self._frontera['este']

        # Resuelve el sistema lineal
        self.sel = Gauss(np.matrix(A), np.matrix(b).transpose(), True)
        self.sel.ajustar_etiquetas(
            ['T_{' + str(i % m + ini_i) + ',' + str(int(i / m) + ini_j) + '}' for i in range(n * m)],
            es_latex=True)
        # Asigna la solución a la matriz U
        self.U[ini_j:n + ini_j, ini_i:m + ini_i] = self.sel.x.reshape(n, m)
        # Calcula los campos de flujo post-proceso
        self._calcular_campos()

    def graficar(self):
        """Genera una gráfica de contorno de la solución.

        Returns
        -------
        Gráfica de los resultados interpolados usando el pquete matplotlib
        """
        plt.axes().set_aspect('equal')
        plt.suptitle('Método de Diferencias Finitas')
        super()._graficar_datos()

    def graficar_campos(self):
        """Genera una gráfica de vectores para los campos de flujo.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib (quiver plot).
        """
        plt.axes().set_aspect('equal')
        plt.suptitle('Método de Diferencias Finitas')
        super()._graficar_campos()


def main():
    """Función principal para demostración."""
    df = DiferenciasFinitas((5, 5), {'norte': 100.0, 'sur': 0.0, 'oeste': 75.0, 'este': 50.0}, k_x=2.0)
    df.graficar()
    df.graficar_valores()
    df.graficar_coordenadas()


if __name__ == '__main__':
    main()



from mnspy.ecuaciones_algebraicas_lineales import EcuacionesAlgebraicasLineales
import numpy as np


class DescomposicionCholesky(EcuacionesAlgebraicasLineales):
    """Clase para la implementación de la descomposición de Cholesky.

    Attributes
    ----------
    _A: ndarray
        matrix cuadrada de coeficientes del sistema
    _b: ndarray
        matrix columna de términos independientes
    _etiquetas: dict[str:list[str], str: bool]
        diccionario con etiquetas de la solución
        key: 'label' contiene una lista de etiquetas de la solución
        key: 'es_latex' un boleano que define si el string es en formato Latex o no
    _n: int
        número de filas de la matrix _A
    _m: int
        número de columnas de la matrix _A
    x: ndarray
        matrix columna de la solución del sistema de acuaciones lineales

    Methods
    -------
    _factorizar():
        Descompone la matrix A en uno matrix triangular superior y una matrix triangular inferior, la cual
        queda guardada en _U.
    retornar_u():
        Retorna la matrix diagonal superior, resultante de la descomposición Cholesky
    retornar_l():
        Retorna la matrix diagonal inferior, resultante de la descomposición Cholesky
    sustituir(b: np.matrix):
        Calcula los resultados de realizar el proceso de sustitución de la matrix b.
        Los resultados quedan guardados en x.
    solucion():
        Presenta los resultados de al realizar el proceso de sustitución. Los resultados quedan guardados en x

    Examples
    -------
    from mnspy import DescomposicionCholesky
    import numpy as np

    A = np.matrix('6 15 55; 15 55 225; 55 225 979')
    print('Matrix A:\n', A)
    ch = DescomposicionCholesky(A)
    print('Matrix superior u:\n', ch.retornar_u())
    print('Matrix inferior l:\n', ch.retornar_l())
    print('Multiplicación u*l:\n', ch.retornar_l() * ch.retornar_u())
    b = np.matrix('76; 295;1 259')
    ch.sustituir(b)
    print('Matrix b:\n', b)
    print('solución:')
    ch.solucion()
    """

    def __init__(self, A: np.ndarray):
        """Constructor de la clase base DescomposicionCholesky

        Parameters
        ----------
        A: ndarray
            Matrix de los coeficientes del sistema de ecuaciones
        """
        b = np.zeros((A.shape[0], 1))
        super().__init__(A, b)
        self._aumentada = None
        self._U = None
        self._factorizar()

    def _factorizar(self):
        """
        Descompone la matrix A en uno matrix triangular superior y una matrix triangular inferior, la cual
        queda guardada en _U

        Returns
        -------
        None
        """
        if self._n != self._m:
            print('La matriz A debe ser cuadrada')
            return
        self._U = np.zeros(self._A.shape)
        for i in range(self._n):
            suma = 0
            for k in range(i):
                suma += self._U[k, i] ** 2
            self._U[i, i] = np.sqrt(self._A[i, i] - suma)
            for j in range(i + 1, self._n):
                suma = 0
                for k in range(i):
                    suma += self._U[k, i] * self._U[k, j]
                self._U[i, j] = (self._A[i, j] - suma) / self._U[i, i]

    def retornar_u(self):
        """
        Retorna la matrix diagonal superior, resultante de la descomposición Cholesky

        Returns
        -------
        objeto ndarray con la matrix diagonal superior
        """
        return self._U

    def retornar_l(self):
        """
        Retorna la matrix diagonal inferior, resultante de la descomposición Cholesky

        Returns
        -------
        objeto ndarray con la matrix diagonal inferior
        """
        return np.transpose(self.retornar_u())

    def sustituir(self, b: np.ndarray):
        """Calcula los resultados de realizar el proceso de sustitución de la matrix b.
        Los resultados quedan guardados en x.

        Parameters
        ----------
        b: ndarray
            Vector columna a la cual se le realizará el proceso de sustitución

        Returns
        -------
        None
        """
        self._aumentada = np.hstack((self.retornar_l(), b))
        d = self.sustitucion_adelante()
        self._aumentada = np.hstack((self.retornar_u(), d))
        self.x = self.sustitucion_atras()


def main():
    A = np.array([[6, 15, 55], [15, 55, 225], [55, 225, 979]])
    print('Matrix A:\n', A)
    ch = DescomposicionCholesky(A)
    print('Matrix superior u:\n', ch.retornar_u())
    print('Matrix inferior l:\n', ch.retornar_l())
    print('Multiplicación u*l:\n', ch.retornar_l() @ ch.retornar_u())
    b = np.array([[76], [295], [1259]])
    ch.sustituir(b)
    print('Matrix b:\n', b)
    print('solución:')
    ch.solucion()


if __name__ == '__main__':
    main()



from mnspy.ecuaciones_diferenciales_parciales.mef import Nodo, Elemento, Rigidez
from mnspy.utilidades import es_notebook, _generar_matrix
from mnspy.ecuaciones_algebraicas_lineales import Gauss
from IPython.display import display, Math
from tabulate import tabulate
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib.transforms as transforms
import sympy as sp

TOL_CERO = 1E-10
sp.init_printing(use_latex=True)

# plt.rcParams.update({
#     "text.usetex": True,
#     "font.family": "sans-serif",
#     "font.sans-serif": "Helvetica",
# })
plt.rcParams.update(plt.rcParamsDefault)


class Viga(Elemento):
    """Representa un elemento de viga que soporta cargas transversales y momentos.

    Este elemento es fundamental para el análisis de vigas y pórticos planos.
    Cada nodo del elemento tiene dos grados de libertad:
    1. Desplazamiento en la dirección 'y' global (transversal).
    2. Rotación alrededor del eje 'z' global (momento flector).

    El elemento puede tener una rótula (articulación) en el nodo final (`nodo_j`),
    lo que significa que no transmite momento en ese punto. Las cargas aplicadas
    se convierten en fuerzas y momentos nodales equivalentes (fuerzas de
    empotramiento perfecto).

    Attributes
    ----------
    _E : float
        Módulo de Young del material.
    _I : float
        Momento de inercia de la sección transversal.
    _k : Rigidez
        Objeto `Rigidez` que contiene la matriz de rigidez del elemento.
    _cargas : dict
        Diccionario que almacena las cargas aplicadas sobre el elemento para
        el cálculo de los diagramas.
    _coef : dict
        Diccionario para almacenar los coeficientes de las ecuaciones de los
        diagramas (constantes de integración).
    """

    def __init__(self, nombre: str, nodo_i: Nodo, nodo_j: Nodo, E: float, I: float):
        """Constructor para el elemento de viga.

        Parameters
        ----------
        nombre : str
            Nombre o identificador del elemento.
        nodo_i : Nodo
            Nodo inicial del elemento.
        nodo_j : Nodo
            Nodo final del elemento. Puede ser una rótula si `nodo_j.es_rotula` es True.
        E : float
            Módulo de Young del material.
        I : float
            Momento de inercia de la sección transversal.
        """
        super().__init__(nombre, nodo_i, nodo_j)

        # --- Matriz de rigidez local [k] ---
        # Se selecciona la matriz de rigidez local apropiada según si el
        # nodo final es una rótula (articulación).
        if nodo_j.es_rotula:
            # Matriz para una viga con apoyo articulado-empotrado (fixed-pinned)
            self._k = Rigidez(np.array([[1, self._L, -1, 0],
                                        [self._L, self._L ** 2, -self._L, 0],
                                        [-1, -self._L, 1, 0],
                                        [0, 0, 0, 0]],
                                       dtype=np.double) * 3 * E * I / self._L ** 3, [self._nodo_i, self._nodo_j],
                              ['y', 'eje_z'])
        else:
            # Matriz para una viga estándar empotrada-empotrada (fixed-fixed)
            self._k = Rigidez(np.array([[12, 6 * self._L, -12, 6 * self._L],
                                        [6 * self._L, 4 * self._L ** 2, -6 * self._L, 2 * self._L ** 2],
                                        [-12, -6 * self._L, 12, -6 * self._L],
                                        [6 * self._L, 2 * self._L ** 2, -6 * self._L, 4 * self._L ** 2]],
                                       dtype=np.double) * E * I / self._L ** 3, [self._nodo_i, self._nodo_j],
                              ['y', 'eje_z'])
        # Vectores de fuerzas nodales equivalentes (fixed-end actions)
        self._fuerzas_i = np.zeros((len(self._k.grados), 1))
        self._fuerzas_j = np.zeros((len(self._k.grados), 1))
        # Vectores para las fuerzas de corrección por rótula
        self._fuerzas_i_rot = np.zeros((len(self._k.grados), 1))
        self._fuerzas_j_rot = np.zeros((len(self._k.grados), 1))

        # Asignación de etiquetas para fuerzas y reacciones en los GDL
        self._nodo_i.grados_libertad['y'].label_reaccion = 'F'
        self._nodo_i.grados_libertad['y'].label_fuerza = 'f'
        self._nodo_i.grados_libertad['eje_z'].label_reaccion = 'M'
        self._nodo_i.grados_libertad['eje_z'].label_fuerza = 'm'
        self._nodo_j.grados_libertad['y'].label_reaccion = 'F'
        self._nodo_j.grados_libertad['y'].label_fuerza = 'f'
        self._nodo_j.grados_libertad['eje_z'].label_reaccion = 'M'
        self._nodo_j.grados_libertad['eje_z'].label_fuerza = 'm'

        # Diccionarios para almacenar cargas y coeficientes de diagramas
        self._cargas = {'distribuida': [], 'puntual': [],
                        'momento': []}
        self._coef = {'A': 0.0, 'B': 0.0, 'c_1': 0.0, 'c_2': 0.0, 'c_3': 0.0, 'c_4': 0.0}
        self._E = E
        self._I = I

    def _repr_latex_(self):
        """Representación en LaTeX del sistema local del elemento para notebooks."""
        vec_d = np.array(self._k.obtener_etiquetas_desplazamientos(), dtype=object).reshape(-1, 1)
        vec_f = np.array(self._obtener_etiquetas_fuerzas(), dtype=object).reshape(-1, 1)
        texto_latex = r'\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(vec_f,
                                       '{:}') + r'\end{array}\right\}_{\{f\}}=\left[\begin{array}{' + 'c' * \
                       self._k.obtener_matriz().shape[1] + '}'
        texto_latex += _generar_matrix(self._k.obtener_matriz(),
                                       '{:G}') + r'\end{array}\right]_{[k]}\cdot\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(vec_d, '{:}') + r'\end{array}\right\}_{\{d\}}-\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(self._obtener_fuerzas(),
                                       '{:}') + r'\end{array}\right\}_{\{f_{o}\}}'
        if self._nodo_j.es_rotula:
            texto_latex += r'-\left\{\begin{array}{c}' + _generar_matrix(self._obtener_fuerzas_por_rotula(),
                                                                         '{:}') + r'\end{array}\right\}_{\{f_{rot}\}}'
        return '$' + texto_latex + '$'

    def __repr__(self):
        """Representación del objeto como string."""
        return 'Viga: ' + self.nombre

    def agregar_carga_distribuida(self, q: float):
        """Agrega una carga uniformemente distribuida.

        Calcula las fuerzas y momentos de empotramiento perfecto (fixed-end actions)
        y los suma a los vectores de fuerza nodal. Si el nodo final es una rótula,
        se aplica una corrección.

        Parameters
        ----------
        q : float
            Magnitud de la carga distribuida. Un valor positivo actúa en la
            dirección +y global.
        """
        # Fuerzas y momentos de empotramiento perfecto para carga distribuida
        self._nodo_i.grados_libertad['y'].fuerza += q * self._L / 2
        self._fuerzas_i[0, 0] += q * self._L / 2
        self._nodo_i.grados_libertad['eje_z'].fuerza += q * self._L ** 2 / 12
        self._fuerzas_i[1, 0] += q * self._L ** 2 / 12
        
        self._nodo_j.grados_libertad['y'].fuerza += q * self._L / 2
        self._fuerzas_j[0, 0] += q * self._L / 2
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q * self._L ** 2 / 12
        self._fuerzas_j[1, 0] += -q * self._L ** 2 / 12

        self._cargas['distribuida'].append([q, q])

        # Si el nodo j es una rótula, se debe liberar el momento en ese extremo.
        # Esto se hace aplicando un momento corrector y las fuerzas cortantes
        # correspondientes para mantener el equilibrio.
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['y'].fuerza -= (-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= (-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-q * self._L ** 2 / 12)
            self._fuerzas_i_rot[1, 0] -= 0.5 * (-q * self._L ** 2 / 12)
            self._nodo_j.grados_libertad['y'].fuerza -= -(-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -(-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q * self._L ** 2 / 12
            self._fuerzas_j_rot[1, 0] -= -q * self._L ** 2 / 12

    def agregar_carga_trapezoidal(self, q_1: float, q_2: float):
        """Agrega una carga trapezoidalmente distribuida."""
        self._nodo_i.grados_libertad['y'].fuerza += q_1 * self._L / 2 + 3 * (q_2 - q_1) * self._L / 20
        self._fuerzas_i[0, 0] += q_1 * self._L / 2 + 3 * (q_2 - q_1) * self._L / 20
        self._nodo_i.grados_libertad['eje_z'].fuerza += q_1 * self._L ** 2 / 12 + (q_2 - q_1) * self._L ** 2 / 30
        self._fuerzas_i[1, 0] += q_1 * self._L ** 2 / 12 + (q_2 - q_1) * self._L ** 2 / 30
        
        self._nodo_j.grados_libertad['y'].fuerza += q_1 * self._L / 2 + 7 * (q_2 - q_1) * self._L / 20
        self._fuerzas_j[0, 0] += q_1 * self._L / 2 + 7 * (q_2 - q_1) * self._L / 20
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20
        self._fuerzas_j[1, 0] += -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20

        self._cargas['distribuida'].append([q_1, q_2])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['y'].fuerza -= (-q_1 * self._L ** 2 / 12 - (
                    q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= (-q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (
                    2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20)
            self._fuerzas_i_rot[1, 0] -= 0.5 * (-q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20)
            self._nodo_j.grados_libertad['y'].fuerza -= -(
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -(-q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (
                    2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20
            self._fuerzas_j_rot[1, 0] -= -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20

    def agregar_carga_puntual(self, p: float, a: float = None):
        """Agrega una carga puntual perpendicular a una distancia 'a'."""
        if a is None:
            a = self._L / 2
        self._nodo_i.grados_libertad['y'].fuerza += p * (self._L - a) ** 2 * (self._L + 2 * a) / self._L ** 3
        self._fuerzas_i[0, 0] += p * (self._L - a) ** 2 * (self._L + 2 * a) / self._L ** 3
        self._nodo_i.grados_libertad['eje_z'].fuerza += p * a * (self._L - a) ** 2 / self._L ** 2
        self._fuerzas_i[1, 0] += p * a * (self._L - a) ** 2 / self._L ** 2
        self._nodo_j.grados_libertad['y'].fuerza += p * a ** 2 * (self._L + 2 * (self._L - a)) / self._L ** 3
        self._fuerzas_j[0, 0] += p * a ** 2 * (self._L + 2 * (self._L - a)) / self._L ** 3
        self._nodo_j.grados_libertad['eje_z'].fuerza += -p * (self._L - a) * a ** 2 / self._L ** 2
        self._fuerzas_j[1, 0] += -p * (self._L - a) * a ** 2 / self._L ** 2

        self._cargas['puntual'].append([p, a])
        self._cargas['puntual'].sort(key=lambda l: l[1])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['y'].fuerza -= (-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= (-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-p * (self._L - a) * a ** 2 / self._L ** 2)
            self._fuerzas_i_rot[1, 0] -= 0.5 * (-p * (self._L - a) * a ** 2 / self._L ** 2)
            self._nodo_j.grados_libertad['y'].fuerza -= -(-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -(-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -p * (self._L - a) * a ** 2 / self._L ** 2
            self._fuerzas_j_rot[1, 0] -= -p * (self._L - a) * a ** 2 / self._L ** 2

    def agregar_momento(self, m: float, a: float = None):
        """Agrega un momento concentrado a una distancia 'a'."""
        m *= -1
        if a is None:
            a = self._L / 2
        self._nodo_i.grados_libertad['y'].fuerza += 6 * m * a * (self._L - a) / self._L ** 3
        self._fuerzas_i[0, 0] += 6 * m * a * (self._L - a) / self._L ** 3
        self._nodo_i.grados_libertad['eje_z'].fuerza += m * (self._L - a) * (3 * a - self._L) / self._L ** 2
        self._fuerzas_i[1, 0] += m * (self._L - a) * (3 * a - self._L) / self._L ** 2
        self._nodo_j.grados_libertad['y'].fuerza += - 6 * m * a * (self._L - a) / self._L ** 3
        self._fuerzas_j[0, 0] += - 6 * m * a * (self._L - a) / self._L ** 3
        self._nodo_j.grados_libertad['eje_z'].fuerza += m * a * (2 * self._L - 3 * a) / self._L ** 2
        self._fuerzas_j[1, 0] += m * a * (2 * self._L - 3 * a) / self._L ** 2

        self._cargas['momento'].append([-m, a])
        self._cargas['momento'].sort(key=lambda l: l[1])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['y'].fuerza -= (m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= (m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (m * a * (2 * self._L - 3 * a) / self._L ** 2)
            self._fuerzas_i_rot[1, 0] -= 0.5 * (m * a * (2 * self._L - 3 * a) / self._L ** 2)
            self._nodo_j.grados_libertad['y'].fuerza -= -(m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -(m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= m * a * (2 * self._L - 3 * a) / self._L ** 2
            self._fuerzas_j_rot[1, 0] -= m * a * (2 * self._L - 3 * a) / self._L ** 2

    def agregar_carga_triangular_descendente(self, q: float):
        """Agrega una carga triangular con valor máximo en el nodo i."""
        self._nodo_i.grados_libertad['y'].fuerza += 7 * q * self._L / 20
        self._fuerzas_i[0, 0] += 7 * q * self._L / 20
        self._nodo_i.grados_libertad['eje_z'].fuerza += q * self._L ** 2 / 20
        self._fuerzas_i[1, 0] += q * self._L ** 2 / 20
        self._nodo_j.grados_libertad['y'].fuerza += 3 * q * self._L / 20
        self._fuerzas_j[0, 0] += 3 * q * self._L / 20
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q * self._L ** 2 / 30
        self._fuerzas_j[1, 0] += -q * self._L ** 2 / 30

        self._cargas['distribuida'].append([q, 0])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['y'].fuerza -= (-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= (-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-q * self._L ** 2 / 30)
            self._fuerzas_i_rot[1, 0] -= 0.5 * (-q * self._L ** 2 / 30)
            self._nodo_j.grados_libertad['y'].fuerza -= -(-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -(-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q * self._L ** 2 / 30
            self._fuerzas_j_rot[1, 0] -= -q * self._L ** 2 / 30

    def agregar_carga_triangular_ascendente(self, q: float):
        """Agrega una carga triangular con valor máximo en el nodo j."""
        self._nodo_i.grados_libertad['y'].fuerza += 3 * q * self._L / 20
        self._fuerzas_i[0, 0] += 3 * q * self._L / 20
        self._nodo_i.grados_libertad['eje_z'].fuerza += q * self._L ** 2 / 30
        self._fuerzas_i[1, 0] += q * self._L ** 2 / 30
        self._nodo_j.grados_libertad['y'].fuerza += 7 * q * self._L / 20
        self._fuerzas_j[0, 0] += 7 * q * self._L / 20
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q * self._L ** 2 / 20
        self._fuerzas_j[1, 0] += -q * self._L ** 2 / 20

        self._cargas['distribuida'].append([0, q])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['y'].fuerza -= (-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= (-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-q * self._L ** 2 / 20)
            self._fuerzas_i_rot[1, 0] -= 0.5 * (-q * self._L ** 2 / 20)
            self._nodo_j.grados_libertad['y'].fuerza -= -(-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -(-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q * self._L ** 2 / 20
            self._fuerzas_j_rot[1, 0] -= -q * self._L ** 2 / 20

    def _obtener_cargas(self) -> dict:
        return self._cargas

    def ecuacion_de_cortante(self):
        x = sp.symbols('x')
        V = sp.Function('V')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        ec = -self._coef['A'] * x ** 2 / 2 - self._coef['B'] * x + self._coef['c_1']
        if len(self._cargas['puntual']) == 0:
            return sp.Eq(V, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in self._cargas['puntual']:
                car.append(car[-1] + c[0])
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(V, sp.Piecewise(*tramos))

    def ecuacion_de_momento(self):
        x = sp.symbols('x')
        M = sp.Function('M')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        ec = -self._coef['A'] * x ** 3 / 6 - self._coef['B'] * x ** 2 / 2 + self._coef['c_1'] * x + self._coef[
            'c_2']
        if len(self._cargas['puntual']) == 0 and len(self._cargas['momento']) == 0:
            return sp.Eq(M, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            cargas_ele = dict()
            for c in self._cargas['puntual']:
                cargas_ele[c[1]] = sp.expand(c[0] * (x - (x_1 + c[1])))
            for m in self._cargas['momento']:
                cargas_ele[m[1]] = sp.expand(cargas_ele.get(m[1], 0.0) - m[0])
            if len(cargas_ele) > 0:
                cargas_ele = dict(sorted(cargas_ele.items(), key=lambda item: item[0]))
            dis = [float(x_1)] + [float(x_1 + c) for c in cargas_ele.keys()] + [float(x_2)]
            # dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in cargas_ele.values():
                car.append(sp.expand(car[-1] + c))
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(M, sp.Piecewise(*tramos))

    def ecuacion_de_giro(self):
        x = sp.symbols('x')
        phi = sp.Function('phi')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        ec = (-self._coef['A'] * x ** 4 / 24 - self._coef['B'] * x ** 3 / 6 + self._coef['c_1'] * x ** 2 / 2 +
              self._coef['c_2'] * x + self._coef['c_3']) / self._E / self._I
        if len(self._cargas['puntual']) == 0 and len(self._cargas['momento']) == 0:
            return sp.Eq(phi, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            cargas_ele = dict()
            for c in self._cargas['puntual']:
                cargas_ele[c[1]] = sp.expand((c[0] * (x - (x_1 + c[1])) ** 2 / 2) / self._E / self._I)
            for m in self._cargas['momento']:
                cargas_ele[m[1]] = sp.expand(cargas_ele.get(m[1], 0.0) - m[0] * (x - (x_1 + m[1])) / self._E / self._I)
            if len(cargas_ele) > 0:
                cargas_ele = dict(sorted(cargas_ele.items(), key=lambda item: item[0]))
            dis = [float(x_1)] + [float(x_1 + c) for c in cargas_ele.keys()] + [float(x_2)]
            # dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in cargas_ele.values():
                car.append(sp.expand(car[-1] + c))
            # for c in self._cargas['puntual']:
            #     car.append(sp.expand(car[-1] + (c[0] * (x - (x_1 + c[1])) ** 2 / 2) / self._E / self._I))
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(phi, sp.Piecewise(*tramos))

    def ecuacion_de_deflexion(self):
        x = sp.symbols('x')
        y = sp.Function('y')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        ec = (-self._coef['A'] * x ** 5 / 120 - self._coef['B'] * x ** 4 / 24 + self._coef['c_1'] * x ** 3 / 6 +
              self._coef['c_2'] * x ** 2 / 2 + self._coef['c_3'] * x + self._coef['c_4']) / self._E / self._I
        x = sp.symbols('x')
        if len(self._cargas['puntual']) == 0 and len(self._cargas['momento']) == 0:
            return sp.Eq(y, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            cargas_ele = dict()
            for c in self._cargas['puntual']:
                cargas_ele[c[1]] = sp.expand((c[0] * (x - (x_1 + c[1])) ** 3 / 6) / self._E / self._I)
            for m in self._cargas['momento']:
                cargas_ele[m[1]] = sp.expand(
                    cargas_ele.get(m[1], 0.0) - (m[0] * (x - (x_1 + m[1])) ** 2 / 2) / self._E / self._I)
            if len(cargas_ele) > 0:
                cargas_ele = dict(sorted(cargas_ele.items(), key=lambda item: item[0]))
            dis = [float(x_1)] + [float(x_1 + c) for c in cargas_ele.keys()] + [float(x_2)]
            # dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in cargas_ele.values():
                car.append(sp.expand(car[-1] + c))

            # car = [ec]
            # for c in self._cargas['puntual']:
            #     car.append(sp.expand(car[-1] + (c[0] * (x - (x_1 + c[1])) ** 3 / 6) / self._E / self._I))
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(y, sp.Piecewise(*tramos))

    def _calcular_coeficientes_diagramas(self):
        """Calcula las constantes de integración para las ecuaciones de los diagramas.

        Resuelve la ecuación diferencial de la viga EIy'''' = w(x) integrando
        cuatro veces. Las cuatro constantes de integración (c1, c2, c3, c4) se
        resuelven utilizando las cuatro condiciones de frontera del elemento:
        los desplazamientos y giros en los nodos i y j.
        """
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        q_1 = q_2 = 0
        for c in self._cargas['distribuida']:
            q_1 += c[0]
            q_2 += c[1]
        q = [0.0, 0.0]
        for c in self._cargas['puntual']:
            q[0] -= c[0] * (self._L - c[1]) ** 3 / 6
            q[1] -= c[0] * (self._L - c[1]) ** 2 / 2
        for m in self._cargas['momento']:
            q[0] -= -m[0] * (self._L - m[1]) ** 2 / 2
            q[1] -= -m[0] * (self._L - m[1])
            
        # Coeficientes de la carga distribuida w(x) = A*x + B
        self._coef['A'] = -(q_2 - q_1) / (x_2 - x_1)
        self._coef['B'] = -q_1 - self._coef['A'] * x_1
        
        # Sistema de ecuaciones para encontrar las constantes de integración [A]{c} = {b}
        A = np.array([[x_1 ** 3 / 6.0, x_1 ** 2 / 2.0, x_1, 1.0], [x_1 ** 2 / 2.0, x_1, 1.0, 0.0],
                      [x_2 ** 3 / 6.0, x_2 ** 2 / 2.0, x_2, 1.0], [x_2 ** 2 / 2.0, x_2, 1.0, 0.0]], dtype=np.double)
        b = self._obtener_desplazamientos() * self._E * self._I
        
        # Añadir términos de las cargas al vector {b}
        b += np.array([[self._coef['A'] * x_1 ** 5 / 120 + self._coef['B'] * x_1 ** 4 / 24],
                       [self._coef['A'] * x_1 ** 4 / 24 + self._coef['B'] * x_1 ** 3 / 6],
                       [self._coef['A'] * x_2 ** 5 / 120 + self._coef['B'] * x_2 ** 4 / 24 + q[0]],
                       [self._coef['A'] * x_2 ** 4 / 24 + self._coef['B'] * x_2 ** 3 / 6 + q[1]]])
        sol = Gauss(A, b, pivote_parcial=True)
        sol.x[abs(sol.x) < TOL_CERO] = 0.0
        self._coef['c_1'] = sol.x[0, 0]
        self._coef['c_2'] = sol.x[1, 0]
        self._coef['c_3'] = sol.x[2, 0]
        self._coef['c_4'] = sol.x[3, 0]

    def _obtener_arrays_cortantes(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_cortante()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_v = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        # lista_p = [[x_1 + c[1], ev(x_1 + c[1])] for c in self._cargas['puntual']]
        lista_p = []
        for c in self._cargas['puntual']:
            for fun in f.args[1].args:
                val, b = fun.subs(x, x_1 + c[1])
                if b:
                    lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_v, lista_p

    def _obtener_arrays_momentos(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_momento()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_m = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        lista_p = [(x_1 + c[1], ev(x_1 + c[1])) for c in self._cargas['puntual']]
        # lista_p = []
        for m in self._cargas['momento']:
            for fun in f.args[1].args:
                val, b = fun.subs(x, x_1 + m[1])
                if b:
                    lista_p.append((x_1 + m[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_m, lista_p

    def _obtener_arrays_angulos(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_giro()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_o = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        lista_p = [[x_1 + c[1], ev(x_1 + c[1])] for c in self._cargas['puntual']]
        lista_p += [(x_1 + m[1], ev(x_1 + m[1])) for m in self._cargas['momento']]
        # lista_p = []
        # for c in self._cargas['puntual']:
        #     for fun in f.args[1].args:
        #         val, b = fun.subs(x, x_1 + c[1])
        #         if b:
        #             lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_o, lista_p

    def _obtener_arrays_deflexion(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_deflexion()
        x_1 = self._nodo_i.punto[0]
        x_2 = self._nodo_j.punto[0]
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_y = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        lista_p = [[x_1 + c[1], ev(x_1 + c[1])] for c in self._cargas['puntual']]
        lista_p += [(x_1 + m[1], ev(x_1 + m[1])) for m in self._cargas['momento']]
        # lista_p = []
        # for c in self._cargas['puntual']:
        #     for fun in f.args[1].args:
        #         val, b = fun.subs(x, x_1 + c[1])
        #         if b:
        #             lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_y, lista_p

    def diagrama_de_cortante(self, n_puntos: int = 500):
        l_x, l_y, l_z = self._obtener_arrays_cortantes(n_puntos)
        plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
        for i in l_z:
            pos_y = 5  # offset escritura
            val_x, val_y = i
            if val_y < 0:
                pos_y = -5
            plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                         textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        plt.grid()
        plt.title('Diagrama de cortante')
        plt.xlabel('$x$')
        plt.ylabel('$V$')
        plt.show()

    def diagrama_de_momento(self, n_puntos: int = 500):
        l_x, l_y, l_z = self._obtener_arrays_momentos(n_puntos)
        plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
        for i in l_z:
            pos_y = 5  # offset escritura
            val_x, val_y = i
            if val_y < 0:
                pos_y = -5
            plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                         textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        plt.grid()
        plt.title('Diagrama de momento')
        plt.xlabel('$x$')
        plt.ylabel('$M$')
        plt.show()

    def diagrama_de_giro(self, n_puntos: int = 500):
        l_x, l_y, l_z = self._obtener_arrays_angulos(n_puntos)
        plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
        for i in l_z:
            pos_y = 5  # offset escritura
            val_x, val_y = i
            if val_y < 0:
                pos_y = -5
            plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                         textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        plt.grid()
        plt.title('Diagrama de giro')
        plt.xlabel('$x$')
        plt.ylabel(r'$\phi$')
        plt.show()

    def diagrama_de_deflexion(self, n_puntos: int = 500):
        l_x, l_y, l_z = self._obtener_arrays_deflexion(n_puntos)
        plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
        for i in l_z:
            pos_y = 5  # offset escritura
            val_x, val_y = i
            if val_y < 0:
                pos_y = -5
            plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                         textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        plt.grid()
        plt.title('Diagrama de deflexión')
        plt.xlabel('$x$')
        plt.ylabel('$y$')
        plt.show()

    def fuerzas_internas(self):
        """Calcula y muestra las fuerzas internas en los nodos del elemento.

        Se utiliza la fórmula {f} = [k]{d} - {f_o}, donde {f_o} es el vector
        de fuerzas nodales equivalentes (fixed-end actions), incluyendo las
        correcciones por rótula si aplica.
        """
        fuerzas_internas = np.matmul(self._k.k,
                                     self._obtener_desplazamientos()) - self._obtener_fuerzas() - self._obtener_fuerzas_por_rotula()
        fuerzas_internas[abs(fuerzas_internas) < TOL_CERO] = 0.0
        if es_notebook():
            indice = ['$' + label + '$' for label in self._obtener_etiquetas_fuerzas()]
            return tabulate({'Fuerzas internas': fuerzas_internas}, headers='keys', showindex=indice, tablefmt='html')
        else:
            print(fuerzas_internas)

    def _obtener_etiquetas_fuerzas(self, reducida: bool = False):
        def ajustar_cadena(s: str) -> str:
            if s == 'eje_z':
                return ''
            else:
                return s

        etq = [nodo.grados_libertad[
                   gl].label_fuerza + '^{(' + self.nombre + ')}_{' + nodo.nombre + ajustar_cadena(gl) + '}'
               for nodo in [self._nodo_i, self._nodo_j] for gl in self._k.grados if
               not reducida or nodo.grados_libertad[gl].valor]
        return etq

    def mostrar_sistema(self, reducida: bool = False):
        """Muestra el sistema de ecuaciones del elemento en formato matricial."""
        vec_d = np.array(self._k.obtener_etiquetas_desplazamientos(reducida), dtype=object).reshape(-1, 1)
        vec_f = np.array(self._obtener_etiquetas_fuerzas(reducida), dtype=object).reshape(-1, 1)
        if es_notebook():
            texto_latex = r'\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_f,
                                           '{:}') + r'\end{array}\right\}_{\{f\}}=\left[\begin{array}{' + 'c' * \
                           self._k.obtener_matriz(reducida).shape[1] + '}'
            texto_latex += _generar_matrix(self._k.obtener_matriz(reducida),
                                           '{:G}') + r'\end{array}\right]_{[k]}\cdot\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_d, '{:}') + r'\end{array}\right\}_{\{d\}}-\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(self._obtener_fuerzas(reducida),
                                           '{:}') + r'\end{array}\right\}_{\{f_{o}\}}'
            display(Math(texto_latex))
        else:
            return np.array2string(self._k.obtener_matriz(reducida),
                                   formatter={'float_kind': lambda x: '{:}'.format(x)})

    def _obtener_desplazamientos(self) -> np.ndarray:
        """Recupera los desplazamientos nodales del elemento desde los nodos."""
        desplazamiento = None
        for item in self._k.lista_nodos:
            d_i = [d.desplazamiento for d in item.grados_libertad.values() if d.gl in self._k.grados]
            if desplazamiento is None:
                desplazamiento = d_i
            else:
                desplazamiento = np.hstack((desplazamiento, d_i))
        b = np.array(desplazamiento).reshape(-1, 1)
        # Si hay una rotula en el nodo derecho, se ajusta el ángulo
        if self._nodo_j.es_rotula:
            b[3] = (self._fuerzas_j[1] * self._L ** 3 / self._E / self._I - np.matmul(np.array(
                [[6 * self._L, 2 * self._L ** 2, - 6 * self._L]]), b[0:3, ])) / 4 / self._L ** 2
        return b

    def __obtener_path_elemento(self, espesor: float) -> Path:
        t = espesor
        x_1, y_1, Z_1 = self._nodo_i.punto
        x_2, y_2, Z_2 = self._nodo_j.punto
        vertices = np.array([(x_1, y_1 - 0.5 * t), (x_2, y_2 - 0.5 * t), (x_2, y_2 + 0.5 * t), (x_1, y_1 + 0.5 * t),
                             (x_1, y_1 - 0.5 * t)])
        codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
        return Path(vertices, codes)

    def diagrama_fuerzas_internas(self):
        """Dibuja un diagrama de cuerpo libre del elemento con sus fuerzas internas."""
        def dibujar_momento(p: tuple[float, float], r: float, teta_1: float = 0, teta_2: float = 90, ejes=None,
                            **kwargs):
            c_x, c_y = p
            r /= 72
            alfa = 0
            teta_flecha = teta_2
            if teta_2 < teta_1:
                teta_1, teta_2 = (teta_2, teta_1)
                alfa = 180
                teta_flecha = teta_1
            trans = (fig.dpi_scale_trans + transforms.ScaledTranslation(c_x, c_y, ejes.transData))
            arc = patches.Arc((0, 0), r, r, angle=0, theta1=teta_1, theta2=teta_2, transform=trans
                              , capstyle='round', linestyle='-', **kwargs)
            ejes.add_patch(arc)
            p_x = (r / 2) * np.cos(np.radians(teta_flecha))
            p_y = (r / 2) * np.sin(np.radians(teta_flecha))
            ejes.add_patch(
                patches.RegularPolygon(xy=(p_x, p_y), numVertices=3, radius=r / 20,
                                       orientation=np.radians(teta_flecha + alfa), transform=trans, **kwargs))

        #############
        # if self._nodo_j.es_rotula:
        #     fo_rotula = np.array(
        #         [[3.0 / (2.0 * self._L)], [0.5], [-3.0 / (2.0 * self._L)]])
        #     fo_rotula = np.vstack((fo_rotula, np.array([[0.0]]))) * self._fuerzas_j[1]
        #     fuerzas_internas = np.matmul(self._k.k,
        #                                  self._obtener_desplazamientos()) - self._obtener_fuerzas() + fo_rotula
        # else:
        #    fuerzas_internas = np.matmul(self._k.k, self._obtener_desplazamientos()) - self._obtener_fuerzas()
        #############
        fuerzas_internas = np.matmul(self._k.k,
                                     self._obtener_desplazamientos()) - self._obtener_fuerzas() - self._obtener_fuerzas_por_rotula()
        fuerzas_internas[abs(fuerzas_internas) < TOL_CERO] = 0.0
        etq = self._obtener_etiquetas_fuerzas()
        fig, ax = plt.subplots()
        ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                       facecolor='lightsteelblue', lw=0.2))
        fuerza = fuerzas_internas[0, 0]
        if fuerza <= 0.0:
            pos = -50
            ali = 'top'
            sentido = '<-'
        else:
            pos = -50
            ali = 'top'
            sentido = '->'
        ax.annotate('$' + etq[0] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(0, pos), textcoords='offset points', va=ali, ha='center', size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[1, 0]
        if fuerza <= 0.0:
            dibujar_momento((self._nodo_i.punto[0], self._nodo_i.punto[1]), 50, 255, 30, ejes=ax, color='b', lw=2)
            pos = -20
            ali = 'right'
        else:
            dibujar_momento((self._nodo_i.punto[0], self._nodo_i.punto[1]), 50, -75, 150, ejes=ax, color='b', lw=2)
            pos = 20
            ali = 'left'
        ax.annotate('$' + etq[1] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos, 20), textcoords='offset points', va='bottom', ha=ali, size=12)
        fuerza = fuerzas_internas[2, 0]
        if fuerza <= 0.0:
            pos = 50
            ali = 'bottom'
            sentido = '->'
        else:
            pos = 50
            ali = 'bottom'
            sentido = '<-'
        ax.annotate('$' + etq[2] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(0, pos), textcoords='offset points', va=ali, ha='center', size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))

        fuerza = fuerzas_internas[3, 0]
        if fuerza <= 0.0:
            dibujar_momento((self._nodo_j.punto[0], self._nodo_j.punto[1]), 50, 255, 30, ejes=ax, color='b', lw=2)
            pos = -20
            ali = 'right'
        else:
            dibujar_momento((self._nodo_j.punto[0], self._nodo_j.punto[1]), 50, -75, 150, ejes=ax, color='b', lw=2)
            pos = 20
            ali = 'left'
        ax.annotate('$' + etq[3] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos, 20), textcoords='offset points', va='bottom', ha=ali, size=12)
        ax.scatter([self._nodo_i.punto[0], self._nodo_j.punto[0]], [self._nodo_i.punto[1], self._nodo_j.punto[1]],
                   c='navy', marker='o')

        ax.axis('equal')
        ax.axis('off')
        ax.set_xmargin(0.15)
        ax.set_ymargin(0.15)
        plt.show()


def main():
    n_1 = Nodo('1', 0, grados_libertad={'y': False, 'eje_z': True})
    n_2 = Nodo('2', 6, grados_libertad={'y': False, 'eje_z': True})

    e_1 = Viga('1', n_1, n_2, E=200E6, I=1E-4)
    e_1.agregar_carga_trapezoidal(-30, -90)
    from mnspy import Ensamble
    mg = Ensamble([e_1])
    mg.diagrama_cargas()
    mg.solucionar_por_gauss_y_calcular_reacciones()
    mg.solucion()
    mg.diagrama_de_cortante()
    mg.diagrama_de_momento()


if __name__ == '__main__':
    main()



"""Módulo Ecuaciones Diferenciales Parciales (EDPs)

Este módulo de `mnspy` proporciona herramientas para resolver ecuaciones
diferenciales parciales (EDPs) elípticas, comúnmente encontradas en problemas
de ingeniería en estado estacionario (ej. Ecuación de Laplace, Ecuación de Poisson).

Submódulos Disponibles
----------------------
- `mdf` (Método de Diferencias Finitas):
    Aproxima las derivadas parciales en una rejilla de puntos (nodos),
    transformando la EDP en un sistema de ecuaciones algebraicas lineales.
    - `DiferenciasFinitas`: Implementa el método de diferencias finitas para mallas rectangulares.
    - `Liebmann`: Solucionador iterativo (Gauss-Seidel con sobrerrelajación) para la ecuación de Laplace.

- `mef` (Método de Elementos Finitos):
    Discretiza el dominio en subdominios más pequeños (elementos) y aproxima
    la solución dentro de cada uno. Es muy flexible para geometrías complejas.

- `mvf` (Método de Volúmenes Finitos):
    Discretiza el dominio en volúmenes de control y aplica las leyes de
    conservación de forma integral sobre cada volumen. Es robusto y garantiza
    la conservación de las propiedades.
"""
from .ecuaciones_diferenciales_parciales import EcuacionesDiferencialesParciales
from .liebmann import Liebmann



"""Módulo de Método de Elementos Finitos (MEF)

Este submódulo de `mnspy` proporciona una implementación orientada a objetos
del Método de Elementos Finitos (MEF o FEM por sus siglas en inglés) para el
análisis estructural de resortes, barras, armaduras, vigas y marcos.

Conceptos Clave
----------------
- **Nodo**: Representa un punto en la estructura donde se conectan los elementos
  y se aplican cargas o restricciones (soportes).
- **Grado de Libertad (GL)**: Describe la capacidad de un nodo para desplazarse
  o rotar en una dirección específica.
- **Elemento**: Es la unidad básica que, al unirse con otras, conforma la
  estructura completa (ej. `Barra`, `Viga`). Cada elemento tiene asociada una
  matriz de rigidez que relaciona fuerzas y desplazamientos.
- **Ensamble**: Es el proceso de combinar las matrices de rigidez de todos los
  elementos para formar la matriz de rigidez global de la estructura,
  permitiendo resolver el sistema de ecuaciones `[K]{d} = {F}`.

Clases Disponibles
------------------
*   `Nodo`, `Elemento`, `GradoLibertad`, `Rigidez`: Clases base.
*   `Resorte`, `Barra`, `Armadura`, `Viga`, `Marco`: Elementos estructurales 1D.
*   `TriangularCST`: Elemento 2D de deformación constante para problemas de tensión plana.
*   `Ensamble`: Clase principal para construir y analizar la estructura completa.
*   Funciones de utilidad como `mallado_estructurado_triangular` e `importar_gmsh`.

"""
from .mef import Nodo, GradoLibertad, Elemento, Rigidez
from .resorte import Resorte
from .barra import Barra
from .armadura import Armadura
from .viga import Viga
from .marco import Marco
from .triangular_cst import TriangularCST
from .ensamble import Ensamble, mallado_estructurado_triangular, importar_gmsh



import matplotlib.pyplot as plt
import numpy as np
plt.rcParams.update(plt.rcParamsDefault)

class EcuacionesDiferencialesParciales:
    """Clase base para los solucionadores de EDPs en mallas estructuradas.

    Esta clase gestiona la creación de la malla, la aplicación de condiciones
    de frontera y los métodos de visualización para problemas de EDPs en
    dominios rectangulares.

    Attributes
    ----------
    U : np.ndarray
        Matriz con los valores de la solución en cada nodo de la malla.
    X : np.ndarray
        Matriz con las coordenadas 'x' de cada nodo.
    Y : np.ndarray
        Matriz con las coordenadas 'y' de cada nodo.
    frontera : dict
        Diccionario que define las condiciones de frontera ('norte', 'sur',
        'este', 'oeste').
    q_x : np.ndarray
        Matriz con los flujos calculados en la dirección x.
    q_y : np.ndarray
        Matriz con los flujos calculados en la dirección y.

    Methods
    -------
    _calcular_campos():
        Calcula los campos secundarios (ej. flujos de calor) a partir de la solución.
    _graficar_datos():
        Genera una gráfica de contorno de la solución.
    graficar_valores():
        Muestra la malla con los valores de la solución en cada nodo.
    graficar_coordenadas():
        Muestra la malla con las coordenadas de cada nodo.
    _graficar_campos():
        Genera una gráfica de vectores (quiver plot) para los campos calculados.
    """
    def __init__(self, n: int | tuple[int, int], frontera: dict[str, float | str | list[float]], val_inicial: float,
                 k_x: float = 1, k_y: float = 1):
        """Constructor de la clase EcuacionesDiferencialesParciales

        Parameters
        ----------
        n: int | tuple[int, int]
            Número de nodos en cada dirección. Si es un entero, la malla será de
            n x n. Si es una tupla (n_filas, n_columnas), la malla será de
            n_filas x n_columnas.
        frontera: dict[str, float | str | list[float]]
            Condiciones de frontera de la placa con las siguientes llaves permitidas:
            'norte', 'sur', 'este', 'oeste'. Los valores pueden ser:
            - ``float``: Condición de Dirichlet (valor constante).
            - ``list[float]``: Condición de Dirichlet (valor variable a lo largo de la frontera).
            - ``str``: Condición de Neumann. Actualmente solo se soporta 'aislado' (flujo cero).
        val_inicial: float
            Valor inicial que tendrá cada uno de los puntos de la placa
        k_x: float, optional
            Coeficiente de conductividad térmica en la dirección x. Por defecto es 1.
        k_y: float, optional
            Coeficiente de conductividad térmica en la dirección y. Por defecto es 1.
        """
        if isinstance(n, tuple):
            self._n, self._m = n
        else:
            self._n = self._m = n
        self._X, self._Y = np.meshgrid(np.arange(0, self._m), np.arange(0, self._n))
        self._v_inicial = val_inicial
        self.U = np.full((self._n, self._m), self._v_inicial, dtype=float)
        self.q_x = np.zeros((self._n, self._m), dtype=float)
        self.q_y = np.zeros((self._n, self._m), dtype=float)
        self._k_x = k_x
        self._k_y = k_y
        self._frontera = frontera
        # Aplicar condiciones de frontera de Dirichlet
        if not self._frontera['norte'] == 'aislado':
            if isinstance(self._frontera['norte'], list):
                self.U[self._n - 1:, :] = np.matrix(self._frontera['norte'])
            else:
                self.U[self._n - 1:, :] = self._frontera['norte']
        if not self._frontera['sur'] == 'aislado':
            if isinstance(self._frontera['sur'], list):
                self.U[:1, :] = np.matrix(self._frontera['sur'])
            else:
                self.U[:1, :] = self._frontera['sur']
        if not self._frontera['oeste'] == 'aislado':
            if isinstance(self._frontera['oeste'], list):
                self.U[:, :1] = np.matrix(self._frontera['oeste']).transpose()
            else:
                self.U[:, :1] = self._frontera['oeste']
        if not self._frontera['este'] == 'aislado':
            if isinstance(self._frontera['este'], list):
                self.U[:, self._m - 1:] = np.matrix(self._frontera['este']).transpose()
            else:
                self.U[:, self._m - 1:] = self._frontera['este']
        plt.ioff()  # deshabilitada interactividad matplotlib

    def _calcular_campos(self):
        """Calcula los campos de flujo (q_x, q_y) a partir de la solución U.

        Utiliza diferencias finitas centradas para aproximar el gradiente.
        """
        for i in range(1, self._n - 1):
            for j in range(1, self._m - 1):
                self.q_x[i, j] = -self._k_x * (self.U[i, j + 1] - self.U[i, j - 1]) / 2
                self.q_y[i, j] = -self._k_y * (self.U[i + 1, j] - self.U[i - 1, j]) / 2

    def _graficar_datos(self):
        """Dibuja un mapa de contorno de la solución `U`.

        Returns
        -------
        Grafica los resultados interpolados, para ello se utiliza el paquete matplotlib
        """
        plt.contourf(self._X, self._Y, self.U, 25, cmap=plt.jet())
        plt.colorbar()
        plt.xlabel('Sur')
        plt.ylabel('Oeste')
        plt.show()

    def graficar_valores(self):
        """Dibuja la malla y muestra el valor de la solución `U` en cada nodo.

        Returns
        -------
        Grafica los valores resultantes en cada punto, para ello se utiliza el paquete matplotlib
        """
        plt.scatter(self._X, self._Y)
        plt.margins(0.1)
        for i in range(self._n):
            for j in range(self._m):
                plt.annotate('{:.2f}'.format(self.U[i, j]), (j, i), xytext=(0, 4), textcoords='offset points',
                         ha='center', va='bottom')
        plt.xlabel('Sur')
        plt.ylabel('Oeste')
        plt.show()

    def graficar_coordenadas(self):
        """Dibuja la malla y muestra las coordenadas `(i, j)` de cada nodo.

        Returns
        -------
        Grafica las coordenadas en cada punto, para ello se utiliza el paquete matplotlib
        """
        plt.scatter(self._X, self._Y)
        plt.margins(0.1)
        for i in range(self._n):
            for j in range(self._m):
                plt.annotate('(' + str(i) + ',' + str(j) + ')', (i, j), xytext=(0, 4), textcoords='offset points',
                         ha='center', va='bottom')
        plt.xlabel('Sur')
        plt.ylabel('Oeste')
        plt.show()

    def _graficar_campos(self):
        """Dibuja un campo de vectores para los flujos (q_x, q_y).

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib (quiver plot).
        """
        plt.quiver(self._X, self._Y, self.q_x, self.q_y, color='g', pivot='mid')
        plt.xlabel('Sur')
        plt.ylabel('Oeste')
        plt.show()



from mnspy.raíces import Raices
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class Muller(Raices):
    """Clase para la implementación del cálculo de raíces por el método abierto de Muller.

    Attributes
    ----------
    f: callable
        Función a la que se le hallará la raíz.
    x: float
        Valor de x inicial.
    h: float
        Valor del delta x que se sumará y restará al x inicial para crear
        los tres puntos iniciales.
    tol: float | int
        Máxima tolerancia del error.
    max_iter: int
        Número máximo de iteraciones permitido.
    tipo_error: str
        Tipo de error a utilizar para la convergencia:
        - ``'%'``: Error relativo porcentual.
        - ``'/'``: Error relativo.
        - ``'n'``: Número de cifras significativas.

    Methods
    -------
    _calcular():
        Realiza los cálculos iterativos del método de Muller.
    graficar(mostrar_sol: bool, mostrar_iter: bool, mostrar_lin_iter: bool, n_puntos: int):
        Genera una gráfica del proceso de búsqueda de la raíz.

    Examples
    -------
    import numpy as np

    def f(x):
        return np.sin(x)

    mu = Muller(f, 5, tol=0.01, tipo_error="%")
    mu.generar_tabla()
    mu.graficar()
    mu.solucion()
    """
    def __init__(self, f: callable, x: float = 0, h: float = 1e-1, tol: float | int = 1e-3, max_iter: int = 20,
                 tipo_error='%'):
        """
        Constructor de la clase Muller.

        Parameters
        ----------
        f: callable
            Función a la que se le hallará la raíz.
        x: float
            Valor de x inicial.
        h: float
            Valor del delta x para generar los puntos iniciales.
        tol: float | int
            Máxima tolerancia del error.
        max_iter: int
            Número máximo de iteraciones permitido.
        tipo_error: str
            Tipo de error a utilizar: ``'%'``, ``'/'`` o ``'n'``.
        """
        # El método de Muller es abierto, por lo que x_min y x_max no se usan en el constructor padre.
        super().__init__(f, None, None, tol, max_iter, tipo_error)
        self.x = x
        self._x_2_i = self._x_0 = self.x
        self._x_1_i = self.x + h
        self._x_0_i = self.x - h
        self._a_0 = []
        self._a_1 = []
        self._a_2 = []
        self._list_x_0 = []
        self._list_x_1 = []
        self._list_x_2 = []
        self._calcular()

    def _calcular(self):
        """Calcula la raíz de la función de forma iterativa usando el método de Muller.

        El método utiliza una parábola que pasa por tres puntos para aproximar
        la siguiente estimación de la raíz.

        Returns
        -------
        None
        """
        # Bucle de iteración para evitar recursión profunda
        while True:
            h_0 = self._x_1_i - self._x_0_i
            h_1 = self._x_2_i - self._x_1_i
            d_0 = (self._f(self._x_1_i) - self._f(self._x_0_i)) / h_0
            d_1 = (self._f(self._x_2_i) - self._f(self._x_1_i)) / h_1

            # Coeficientes de la parábola a*x^2 + b*x + c
            a = (d_1 - d_0) / (h_1 + h_0)
            b = a * h_1 + d_1
            c = self._f(self._x_2_i)

            # Guardar coeficientes para graficar
            self._a_2.append(a)
            self._a_1.append(d_0)
            self._a_0.append(self._f(self._x_0_i))
            self._list_x_0.append(self._x_0_i)
            self._list_x_1.append(self._x_1_i)
            self._list_x_2.append(self._x_2_i)

            # Fórmula cuadrática, se elige el denominador más grande para evitar restas catastróficas.
            rad = np.sqrt(b ** 2 - 4 * a * c)
            if abs(b + rad) > abs(b - rad):
                den = b + rad
            else:
                den = b - rad

            self.x = self._x_2_i - (2 * c / den)

            if self._fin_iteracion():
                break  # Convergencia o máximo de iteraciones alcanzado

            # Actualizar los puntos para la siguiente iteración
            self._x_0_i, self._x_1_i, self._x_2_i = self._x_1_i, self._x_2_i, self.x

    def graficar(self, mostrar_sol: bool = True, mostrar_iter: bool = True, mostrar_lin_iter: bool = True,
                 n_puntos: int = 100):
        """Genera una gráfica del proceso de búsqueda de la raíz por el método de Muller.

        Parameters
        ----------
        mostrar_sol: bool
            Si es ``True``, muestra el punto de la solución final.
        mostrar_iter: bool
            Si es ``True``, muestra los puntos de cada iteración.
        mostrar_lin_iter: bool, optional
            Si es ``True``, muestra las parábolas de interpolación utilizadas en
            cada iteración. Por defecto es ``True``.
        n_puntos: int, optional
            Número de puntos para dibujar la curva de la función. Por defecto es 100.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        plt.autoscale(False)
        if mostrar_lin_iter and len(self._tabla['x']) > 0:
            x = np.linspace(min(self._tabla['x'] + [self._x_0]), max(self._tabla['x'] + [self._x_0]), n_puntos)
            for i in range(len(self._tabla['x'])):
                y = self._a_2[i] * (x - self._list_x_0[i]) * (x - self._list_x_1[i]) + self._a_1[i] * (
                        x - self._list_x_0[i]) + self._a_0[i]
                plt.plot(x, y, linestyle='dashed', c='purple', lw=1)
                plt.plot([self._tabla['x'][i], self._tabla['x'][i]], [0, self._f(self._tabla['x'][i])], linestyle='dashed',
                         c='purple', lw=1)
        plt.autoscale(True)
        plt.title(f'x = {self.x}')
        if self._error_porcentual:
            plt.suptitle('Método de Müller (tol= ' + str(self._tol) + '%)')
        else:
            plt.suptitle('Método de Müller (tol= ' + str(self._tol) + ')')
        super().graficar(mostrar_sol, mostrar_iter, mostrar_lin_iter, n_puntos)


def main():
    """Función principal para demostración."""
    def f(x):
        # return x ** 3 - 13 * x - 12
        return np.sin(x)

    mu = Muller(f, 5, tol=0.01, tipo_error="%")
    mu.generar_tabla()
    mu.graficar()
    mu.solucion()


if __name__ == '__main__':
    main()



"""
mnspy: Métodos Numéricos y Simulación en Python.

`mnspy` es una biblioteca de Python desarrollada con fines didácticos para facilitar
la enseñanza y el aprendizaje de métodos numéricos y simulación a estudiantes
de ingeniería.

La biblioteca implementa algoritmos numéricos fundamentales utilizando un enfoque
orientado a objetos, lo que permite una comprensión más clara de los conceptos
teóricos.

Módulos Principales
-------------------
La biblioteca se organiza en los siguientes módulos, cada uno enfocado en un área
específica de los métodos numéricos:

- `raíces`:
    Métodos para encontrar las raíces de ecuaciones no lineales.
- `ecuaciones_algebraicas_lineales`:
    Algoritmos para resolver sistemas de ecuaciones lineales.
- `interpolación`:
    Técnicas de interpolación polinómica.
- `integrales`:
    Métodos para la integración numérica de funciones.
- `derivada`:
    Cálculo numérico de derivadas.
- `ecuaciones_diferenciales_ordinarias` (EDOs):
    Solucionadores para ecuaciones diferenciales ordinarias.
- `ecuaciones_diferenciales_parciales` (EDPs):
    Solucionadores para ecuaciones diferenciales parciales, incluyendo:
    - Método de Diferencias Finitas (`mdf`)
    - Método de Elementos Finitos (`mef`)
    - Método de Volúmenes Finitos (`mvf`)
- `utilidades`:
    Funciones y clases auxiliares utilizadas en toda la biblioteca.

`mnspy` está diseñado para ser una herramienta de apoyo que permita a los
estudiantes visualizar y experimentar con los algoritmos, reforzando así su
comprensión de los principios de la simulación científica.
"""
from .derivada import *
from .ecuaciones_diferenciales_ordinarias import *
from .ecuaciones_diferenciales_parciales import *
from .ecuaciones_diferenciales_parciales.mdf import *
from .ecuaciones_diferenciales_parciales.mef import *
from .ecuaciones_diferenciales_parciales.mvf import *
from .ecuaciones_algebraicas_lineales import *
from .integrales import *
from .interpolación import *
from .raíces import *
from .utilidades import *



"""Módulo Interpolación - Métodos de Ajuste de Curvas

Este módulo de `mnspy` proporciona una colección de métodos para la
interpolación de datos. La interpolación es una técnica para estimar valores
desconocidos que se encuentran entre puntos de datos conocidos.

Se incluyen tanto métodos polinómicos globales (que usan todos los puntos para
crear un único polinomio) como métodos de trazadores (splines), que ajustan
polinomios por tramos entre los puntos.

Clases Disponibles
------------------
- `Interpolacion`: Clase base para todos los métodos de interpolación.
- `InterpolacionNewton`: Interpolación polinómica de Newton (diferencias divididas).
- `InterpolacionLagrange`: Interpolación polinómica de Lagrange.
- `SplineLineal`: Trazadores lineales que conectan puntos con rectas.
- `SplineCubica`: Trazadores cúbicos que aseguran la suavidad en las uniones.

"""
from .interpolacion import Interpolacion
from .inter_Newton import InterpolacionNewton
from .inter_Lagrange import InterpolacionLagrange
from .inter_spline_lineal import SplineLineal
from .inter_spline_cubica import SplineCubica



from mnspy.utilidades import es_notebook, mostrar_matrix, _generar_matrix
import matplotlib.pyplot as plt
import numpy as np
from tabulate import tabulate
from IPython.display import display, Math

plt.rcParams.update(plt.rcParamsDefault)


class EcuacionesAlgebraicasLineales:
    """
    Clase base para las demás clases que implementan métodos para el cálculo de ecuaciones algebraícas lineales.

    Attributes
    ----------
    _A: ndarray
        matrix cuadrada de coeficientes del sistema
    _b: ndarray
        matrix columna de términos independientes
    _aumentada: ndarray
        matrix combinanada para solución del sistema de ecuaciones lineales
    _etiquetas: dict[str:list[str], str: bool]
        diccionario con etiquetas de la solución
        key: 'label' contiene una lista de etiquetas de la solución
        key: 'es_latex' un boleano que define si el string es en formato Latex o no
    _n: int
        número de filas de la matrix _A
    _m: int
        número de columnas de la matrix _A
    _pivote: ndarray
        Solo se usa en la Descomposición LU y es una matrix cuadrada de pivotes
    x: ndarray
        matrix columna de la solución del sistema de acuaciones lineales

    Methods
    -------
    graficar():
        presenta la solución gráfica del sistema de ecuaciones, solo aplica para un sistema de 2x2

    eliminacion_gauss_jordan():
        ejecuta la eliminación de Gauss-Jordan a la matrix self._aumentada

    eliminacion_adelante():
        ejecuta la eliminación hacia adelante a la matrix self._aumentada, usada en el método de Gauss

    sustitucion_atras():
        ejecuta la sustitución hacia atrás a la matrix self._aumentada, usada en el método de Gauss

    sustitucion_adelante():
        ejecuta la sustitución hacia adelante a la matrix self._aumentada, usada en los métodos de descomposición
        de Cholesky y en la descomposición LU

    solucion():
        Presenta los resultados de la solución del sistema de ecuaciones

    Examples:
    -------
    from mnspy import EcuacionesAlagebraicasLineales
    from mnspy.utilidades import mostrar_matrix
    import numpy as np

    A = np.matrix('3 2;-1 2')
    mostrar_matrix(A)
    b = np.matrix('18 ; 2')
    sol = EcuacionesAlagebraicasLineales(A, b)
    sol.mostrar_A()
    sol.mostrar_b()
    sol.graficar()
    sol.mostrar_sistema()
    """

    def __init__(self, A: np.ndarray, b: np.ndarray):
        """Constructor de la clase base EcuacionesAlagebraicasLineales

        Parameters
        ----------
        A: ndarray
            Matrix de los coeficientes del sistema de ecuaciones
        b: ndarray
            Vector columna de los terminos independiantes
        """
        self._A = A.astype(np.double)
        self._b = b.astype(np.double)
        self._n, self._m = self._A.shape  # Se obtiene las dimensiones
        self.x = np.zeros(self._b.shape)
        # self.x = np.matrix(self.x)
        self._aumentada = np.hstack((self._A, self._b))
        self._etiquetas = None
        self._pivote = np.eye(self._n)  # Solo se usa en la Descomposición LU
        plt.ioff()  # deshabilitada interactividad matplotlib

    def graficar(self, x_min: float = None, x_max: float = None, max_rango: float = 10.0):
        """

        Parameters
        ----------
        x_min : float
            valor mínimo en x de la gráfica, por defecto es None
        x_max : float
            valor máximo en x de la gráfica, por defecto es None
        max_rango : float
            si alguno de los valores x_min o x_max es None, el rango en x de la gráfica sería max_rango con
            la solución en el centro

        Returns
        -------
        gráfica usando el paquete matplotlib
        """
        if self._A.shape[0] == 2 and self._A.shape[1] == 2:
            def f1(var):
                return -(self._A[0, 0] / self._A[0, 1]) * var + self._b[0, 0] / self._A[0, 1]

            def f2(var):
                return -(self._A[1, 0] / self._A[1, 1]) * var + self._b[1, 0] / self._A[1, 1]

            singular = False
            div = self._A[1, 0] / self._A[1, 1] - self._A[0, 0] / self._A[0, 1]
            if div != 0:
                sol_x = (self._b[1, 0] / self._A[1, 1] - self._b[0, 0] / self._A[0, 1]) / div
            else:
                sol_x = 0
                singular = True
            sol_y = f1(sol_x)
            if x_min is None or x_max is None:
                x_min = sol_x - max_rango / 2
                x_max = sol_x + max_rango / 2
            x = np.linspace(x_min, x_max, 100)
            y_1 = f1(x)
            y_2 = f2(x)
            plt.plot(x, y_1, c='b', lw=2, label='$f_{1}(x)$')
            plt.plot(x, y_2, c='g', lw=2, label='$f_{2}(x)$')
            if not singular:
                plt.scatter(sol_x, sol_y, c='r', lw=2, label='Solución')
                plt.annotate('$(' + str(sol_x) + r',\,' + str(sol_y) + ')$', (sol_x, sol_y), c='r', alpha=0.9,
                             textcoords="offset points",
                             xytext=(0, 10), ha='center')
            plt.grid()
            plt.legend()
            plt.show()
        else:
            print('Solo gráfica sistemas de ecuaciones de 2 x 2')

    def eliminacion_gauss_jordan(self):
        """
        aplica el algoritmo de Gauss-Jordan sobre la matrix _aumentada

        Returns
        -------
        None
        """
        for k in range(self._n):
            self._pivoteo_parcial(k)
            # Normalización de la fila
            self._aumentada[k, k:] /= self._aumentada[k, k]
            for i in range(self._n):
                if i != k:
                    self._aumentada[i, k:] -= self._aumentada[i, k] * self._aumentada[k, k:]

    def eliminacion_adelante(self, pivote_parcial: bool = False, guardar_factores: bool = False):
        """Ejecuta la eliminación hacia adelante a la matrix _aumentada

        Parameters
        ----------
        pivote_parcial: bool
            Si es verdadero aplica elimicación hacia adelante de Gauss con pivote parcial a la matrix _aumentada,
            en caso de ser falso no aplica el pivote parcial
        guardar_factores: bool
            Si es verdadero aplica eliminación hacia adelante de Gauss a la matrix _aumentada, guardando los valores
            en la matrix, es usada por la Descomposición LU, en caso de ser falso no guarda los valores
        Returns
        -------
        None
        """
        for k in range(self._n - 1):
            if pivote_parcial:
                self._pivoteo_parcial(k)
            for i in range(k + 1, self._n):
                factor = self._aumentada[i, k] / self._aumentada[k, k]
                self._aumentada[i, k:] = self._aumentada[i, k:] - factor * self._aumentada[k, k:]
                if guardar_factores:
                    self._aumentada[i, k] = factor

    def sustitucion_atras(self) -> np.matrix | np.ndarray:
        """Ejecuta la sustitució hacia atrás a la matrix _aumentada

        Returns
        -------
        ndarray que corresponde al vector columna del resultado de la sustitución hacia atrás
        """
        x = np.zeros([self._n, 1])  # Se crea array en cero
        # x = np.matrix(x)  # Se convierte a matrix
        x[self._n - 1] = self._aumentada[self._n - 1, self._n] / self._aumentada[self._n - 1, self._n - 1]
        for i in range(self._n - 2, -1, -1):
            x[i] = (self._aumentada[i, self._n] - np.matmul(self._aumentada[[i], i + 1:self._n],
                                                            x[i + 1:self._n, [0]])) / \
                   self._aumentada[i, i]
        return x

    def sustitucion_adelante(self) -> np.matrix | np.ndarray:
        """Ejecuta la sustitució hacia adelante a la matrix _aumentada

        Returns
        -------
        ndarray que corresponde al vector columna del resultado de la sustitución hacia adelante
        """
        x = np.zeros([self._n, 1])  # Se crea array en cero
        # x = np.matrix(x)  # Se convierte a matrix
        # el término aii=1 para LU
        x[0] = self._aumentada[0, self._n] / self._aumentada[0, 0]
        for i in range(1, self._n):
            x[i] = (self._aumentada[i, self._n] - np.matmul(self._aumentada[[i], 0:i + 1], x[0:i + 1, [0]])) / \
                   self._aumentada[i, i]
        return x

    def _pivoteo_parcial(self, fila: int) -> None:
        """Realiza un pivoteo parcial al número de la fila indicada en la matrix _aumentada, busca hacia adelante
        la fila que contiene el mayor valor abasoluto para esa diagonal y la intercambia.

        Parameters
        ----------
        fila: int
            Fila a la que se le realizará el pivoteo parcial.

        Returns
        -------
        None
        """
        for piv in range(fila + 1, self._n):
            if abs(self._aumentada[piv, fila]) > abs(self._aumentada[fila, fila]):
                self._aumentada[[piv, fila]] = self._aumentada[[fila, piv]]  # Intercambio de filas
                self._pivote[[piv, fila]] = self._pivote[[fila, piv]]  # Intercambio de filas pivote

    def solucion(self):
        """Muestra la solución del sistema de ecuaciones lineales de acuerdo a las etiquetas definidas

        Returns
        -------
        Tabla de resultados usando el paquete tabulate
        """
        if es_notebook():
            if self._etiquetas is None:
                indice = ['$x_{' + str(i) + '}$' for i in range(self.x.shape[0])]
            else:
                if self._etiquetas['es_latex']:
                    indice = ['$' + label + '$' for label in self._etiquetas['label']]
                else:
                    indice = self._etiquetas['label']
            return tabulate({'Solución': self.x}, headers='keys', showindex=indice, tablefmt='html')
        else:
            if self._etiquetas is None:
                indice = ['x_' + str(i) for i in range(self.x.shape[0])]
            else:
                indice = self._etiquetas['label']
            print(tabulate({'Solución': self.x}, headers='keys', showindex=indice, tablefmt='simple'))

    def ajustar_etiquetas(self, etiquetas: list, es_latex: bool = True):
        """Ajusta el nombre de las etiquetas de la solución

        Parameters
        ----------
        etiquetas: list[str]
            Diccionario con etiquetas de la solución
        es_latex: bool
            Un boleano que define si el string es en formato Latex o no

        Returns
        -------
        None
        """
        self._etiquetas = {'label': etiquetas, 'es_latex': es_latex}

    def mostrar_aumentada(self, n_decimal=None):
        """Muestra la matrix aumentada

        Parameters
        ----------
        n_decimal: int
            Número de decimales que tendrá los elementos de la matrix

        Returns
        -------
        Render de la matrix en Latex si se encuentra en un notebook de jupyter o en caso contrario mostrará la
        impresión de la matrix
        """
        i, j = self._aumentada.shape
        return mostrar_matrix(self._aumentada, n_decimal, j - i)

    def mostrar_A(self, n_decimal=None):
        """Muestra la matrix de coeficientes A

        Parameters
        ----------
        n_decimal: int
            Número de decimales que tendrá los elementos de la matrix

        Returns
        -------
        Render de la matrix en Latex si se encuentra en un notebook de jupyter o en caso contrario mostrará la
        impresión de la matrix
        """
        return mostrar_matrix(self._A, n_decimal)

    def mostrar_b(self, n_decimal=None):
        """Muestra la matrix de terminos independientes b

        Parameters
        ----------
        n_decimal: int
            Número de decimales que tendrá los elementos de la matrix

        Returns
        -------
        Render de la matrix en Latex si se encuentra en un notebook de jupyter o en caso contrario mostrará la
        impresión de la matrix
        """
        return mostrar_matrix(self._b, n_decimal)

    def mostrar_sistema(self, n_decimal=None):
        """Muestra el sistema de ecuaciones formado en forma matricial

        Parameters
        ----------
        n_decimal: int
            Número de decimales que tendrá los elementos de la matrix

        Returns
        -------
        Render del sistema de ecuaciones formado en forma matricial en Latex si se encuentra en un notebook de jupyter
        o en caso contrario mostrará la impresión del sistema de ecuaciones
        """
        if n_decimal is None:
            fmt = '{:G}'
        else:
            fmt = '{:.' + str(n_decimal) + 'f}'
        if self._etiquetas is None:
            vec_x = np.array(['x_{' + str(i) + '}' for i in range(self.x.shape[0])], dtype=object).reshape(-1, 1)
            vec_x_plano = np.array(['x_' + str(i) for i in range(self.x.shape[0])], dtype=object).reshape(-1, 1)
        else:
            if self._etiquetas['es_latex']:
                vec_x = np.array([label for label in self._etiquetas['label']], dtype=object).reshape(-1, 1)
            else:
                vec_x = np.array([r'\textrm{' + label + '}' for label in self._etiquetas['label']],
                                 dtype=object).reshape(-1, 1)
            vec_x_plano = np.array([label for label in self._etiquetas['label']], dtype=object).reshape(-1, 1)
        texto_latex = r'\left[\begin{array}{' + 'c' * self._A.shape[1] + '}'
        if es_notebook():
            texto_latex += _generar_matrix(self._A, fmt) + r'\end{array}\right]\cdot\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_x, '{:}') + r'\end{array}\right\}=\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(self._b, fmt) + r'\end{array}\right\}'
            return display(Math(texto_latex))
        else:
            print(np.array2string(self._A, formatter={'float_kind': lambda x: fmt.format(x)}))
            print('x')
            print(vec_x_plano)
            print('=')
            print(np.array2string(self._b, formatter={'float_kind': lambda x: fmt.format(x)}))

    def mostrar_determinantes(self, n_decimal=None):
        """Muestra los determinantes del sistema y el de cada variable (método de cramer), solo para un notebook de jupyter

        Parameters
        ----------
        n_decimal: int
            Número de decimales que tendrá los elementos de la matrix

        Returns
        -------
        Render del sistema de los determinantes en Latex si se encuentra en un notebook de jupyter, en caso contrario
        no se mostrará nada
        """
        if n_decimal is None:
            fmt = '{:}'
        else:
            fmt = '{:.' + str(n_decimal) + 'g}'
        if es_notebook():
            det = np.linalg.det(self._A)
            texto_latex = r'\begin{flalign}\Delta S=\begin{vmatrix}'
            texto_latex += _generar_matrix(self._A, fmt) + r'\end{vmatrix}&=' + fmt.format(det) + r'&&\\'
            for i in range(self._A.shape[1]):
                if self._etiquetas is None:
                    texto_latex += r'\Delta x_{' + str(i) + r'}=\begin{vmatrix}'
                else:
                    texto_latex += r'\Delta ' + self._etiquetas['label'][i] + r'=\begin{vmatrix}'
                mat_i = self._A.copy()
                mat_i[:, [i]] = self._b
                texto_latex += _generar_matrix(mat_i, fmt) + r'\end{vmatrix}&=' + fmt.format(
                    np.linalg.det(mat_i)) + r'&&\\'
            for i in range(self._A.shape[1]):
                if self._etiquetas is None:
                    texto_latex += 'x_{' + str(i) + r'}=\dfrac{\Delta x_{' + str(i) + r'}}{\Delta S}&=' + fmt.format(
                        self.x[i, 0]) + r'&&\\'
                else:
                    texto_latex += self._etiquetas['label'][i] + r'=\dfrac{\Delta ' + self._etiquetas['label'][
                        i] + r'}{\Delta S}&=' + fmt.format(self.x[i, 0]) + r'&&\\'
            texto_latex += r'\end{flalign}'
            return display(Math(texto_latex))
        else:
            return None


def main():
    A = np.array([[3, 2], [-1, 2]])
    mostrar_matrix(A)
    b = np.array([[18], [2]])
    sol = EcuacionesAlgebraicasLineales(A, b)
    sol.mostrar_A()
    sol.mostrar_b()
    sol.graficar()
    sol.mostrar_sistema()


if __name__ == '__main__':
    main()



from mnspy.raíces import Raices
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class SecanteModificada(Raices):
    """Clase para la implementación del cálculo de raíces por el método abierto de la Secante Modificada.

    Attributes
    ----------
    f: callable
        Función a la que se le hallará la raíz.
    x: float
        Valor de x inicial.
    delta: float
        Pequeña fracción de x utilizada para aproximar la derivada.
    tol: float | int
        Máxima tolerancia del error.
    max_iter: int
        Número máximo de iteraciones permitido.
    tipo_error: str
        Tipo de error a utilizar para la convergencia:
        - ``'%'``: Error relativo porcentual.
        - ``'/'``: Error relativo.
        - ``'n'``: Número de cifras significativas.

    Methods
    -------
    _calcular():
        Realiza los cálculos iterativos del método de la Secante Modificada.
    graficar(mostrar_sol: bool, mostrar_iter: bool, mostrar_lin_iter: bool, n_puntos: int):
        Genera una gráfica del proceso de búsqueda de la raíz.

    Examples
    -------
    from mnspy import SecanteModificada

        def f(x):
        return x ** 3 - 10 * x ** 2 + 5

    se = SecanteModificada(f, 0.2, delta=0.1, tol=0.01, tipo_error="%")
    se.generar_tabla()
    se.graficar()
    se.solucion()
    """
    def __init__(self, f: callable, x: float = 0, delta: float = 1e-6, tol: float | int = 1e-3, max_iter: int = 20,
                 tipo_error='%'):
        """
        Constructor de la clase SecanteModificada.

        Parameters
        ----------
        f: callable
            Función a la que se le hallará la raíz.
        x: float
            Valor de x inicial.
        delta: float
            Pequeña fracción de x para aproximar la derivada.
        tol: float | int
            Máxima tolerancia del error.
        max_iter: int
            Número máximo de iteraciones permitido.
        tipo_error: str
            Tipo de error a utilizar: ``'%'``, ``'/'`` o ``'n'``.
        """
        # El método es abierto, no requiere x_min y x_max.
        super().__init__(f, None, None, tol, max_iter, tipo_error)
        self._x_0 = self.x = x
        self._delta = delta
        self._calcular()

    def _calcular(self):
        """Calcula la raíz de la función de forma iterativa usando el método de la Secante Modificada.

        Este método aproxima la derivada utilizando una pequeña perturbación `delta`,
        similar a una diferencia finita.

        Returns
        -------
        None
        """
        self.x -= (self._f(self.x) * self._delta) / (
                self._f(self.x + self._delta) - self._f(self.x))
        if self._fin_iteracion():
            return
        else:
            self._calcular()

    def graficar(self, mostrar_sol: bool = True, mostrar_iter: bool = True, mostrar_lin_iter: bool = True,
                 n_puntos: int = 100):
        """Genera una gráfica del proceso de búsqueda de la raíz por el método de la Secante Modificada.

        Parameters
        ----------
        mostrar_sol: bool, optional
            Si es ``True``, muestra el punto de la solución final. Por defecto es ``True``.
        mostrar_iter: bool, optional
            Si es ``True``, muestra los puntos de cada iteración. Por defecto es ``True``.
        mostrar_lin_iter: bool
            Si es ``True``, muestra las líneas secantes utilizadas en cada iteración. Por defecto es ``True``.
        n_puntos: int, optional
            Número de puntos para dibujar la curva de la función. Por defecto es 100.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        if mostrar_lin_iter and len(self._tabla['x']) > 0:
            # fun = vectorize(self.f)
            l_x = self._tabla['x'].copy()
            l_x.insert(0, self._x_0)
            for i, x in enumerate(l_x[:-1]):
                y = self._f(x)
                x_next = l_x[i + 1]
                derivada_aprox = (self._f(x + self._delta) - self._f(x)) / self._delta
                plt.axline((x, y), slope=derivada_aprox, linestyle='dashed',
                           c='purple',
                           lw=1)
                plt.plot([x_next, x_next], [0, self._f(x_next)], linestyle='dashed', c='purple', lw=1)
        plt.title(f'x = {self.x}')
        if self._error_porcentual:
            plt.suptitle(
                r'Método de la Secante Modificada, $\delta$ = ' + str(self._delta) + ' (tol= ' + str(self._tol) + '%)')
        else:
            plt.suptitle(
                r'Método de la Secante Modificada, $\delta$ = ' + str(self._delta) + ' (tol= ' + str(self._tol) + ')')
        super().graficar(mostrar_sol, mostrar_iter, mostrar_lin_iter, n_puntos)


def main():
    """Función principal para demostración."""
    def f(x):
        return x ** 3 - 10 * x ** 2 + 5

    se = SecanteModificada(f, 0.2, delta=0.1, tol=0.01, tipo_error="%")
    se.generar_tabla()
    se.graficar()
    se.solucion()


if __name__ == '__main__':
    main()
